import collections
import re
import math
import pandas as pd
import spacy
from Levenshtein import distance as lev


def get_entities(doc, labels):
    entities = []
    for ent in doc.ents:
        if ent.label_ in labels:
            entities.append(ent)
    return entities


def calculate_lev(names, threshold):
    pairs = {}
    deselect = []
    for i, name in enumerate(names):
        if i in deselect:
            continue
        pair = []

        for j in range(i + 1, len(names)):
            dis = lev(name, names[j])
            if dis <= threshold:
                pair.append(j)
                deselect.append(j)
        pairs[i] = pair

    return pairs, len(pairs.keys())


def get_supra_match_by_fuzzy(supra,precedent_breakup,threshold):
        matches=[]
        supra_text = re.sub(' +', '', supra.text.lower())

        for i, precedent in enumerate(precedent_breakup.keys()):
            pet = precedent_breakup[precedent][0]
            res = precedent_breakup[precedent][1]
            pet_text=''
            res_text=''
            if precedent.start > supra.end:
                break

            if pet != None:
                pet = pet.split(' and ')[0]
                pet_text = re.sub(' +', ' ', pet.lower())

            if res != None:
                res = res.split(' and ')[0]
                res_text = re.sub(' +', ' ', res.lower())

            supra_text = supra_text.replace('(', r'\(').replace(')', r'\)')

            if pet == None and res == None:
                continue

            found = 0
            req_length=len(supra_text.split(' '))

            for petitioner in pet_text.split(' '):
                for s in supra_text.split(' '):
                    if lev(petitioner, s) <= threshold:
                        found = found + 1
                        threshold=threshold-lev(petitioner,s)

            if found != req_length:


                for respondent in res_text.split(' '):
                    for s in supra_text.split(' '):
                        if lev(respondent, s) <= threshold:
                            found = found + 1
                            threshold = threshold - lev(respondent, s)

                            # matches.append(precedent)




            if found==req_length:

                    matches.append(precedent)


        if len(matches) > 0:
            return matches[-1]
        else:
            return None

def get_precedent_supras(doc, entities_pn, precedent_breakup, entities_precedents, changes_threshold=2):
    text = doc.text
    ends = [ent.end_char for ent in entities_pn]
    supras = []
    for match in re.finditer(r'(\'s\s*case\s*\(supra\)|\s*\(supra\))', text):

        if match.start() in ends:

            supras.append(entities_pn[ends.index(match.start())])
        elif match.start() - 1 in ends:
            supras.append(entities_pn[ends.index(match.start() - 1)])

    supra_precedent_matches = {}

    for supra in supras:
        matches = []
        supra_text = re.sub(' +', '', supra.text)

        for i, precedent in enumerate(entities_precedents):
            if precedent.start > supra.end:
                break

            precedent_text = re.sub(' +', '', precedent.text)
            supra_text = supra_text.replace('(', r'\(').replace(')', r'\)')
            match = re.search(supra_text, precedent_text, re.IGNORECASE)

            if match:
                matches.append(precedent)
        if len(matches) > 0:
            supra_precedent_matches[supra] = matches[-1]
        else:
            match=get_supra_match_by_fuzzy(supra,precedent_breakup,changes_threshold)

            if match != None:
                supra_precedent_matches[supra]=match


    return supra_precedent_matches, supras


def create_precedent_clusters(precedent_breakup, percentage_threshold):
    cluster_num = 0
    exclude = []
    precedent_clusters = {}

    for i, pre in enumerate(precedent_breakup.keys()):

        if i in exclude:
            continue
        pet = precedent_breakup[pre][0]
        res = precedent_breakup[pre][1]
        cit = precedent_breakup[pre][2]
        threshold_pet=math.ceil((percentage_threshold/100)*len(pet))
        threshold_res=math.ceil((percentage_threshold/100)*len(res))

        cluster = []
        cluster.append(pre)
        if pet != None and res != None:

            for j in range(i + 1, len(precedent_breakup)):

                pet_1 = list(precedent_breakup.values())[j][0]
                res_1 = list(precedent_breakup.values())[j][1]
                cit_1 = list(precedent_breakup.values())[j][2]
                if (pet_1 == None or res_1 == None):
                    if cit_1 == None:
                        exclude.append(j)

                    else:
                        if cit_1 == cit:
                            exclude.append(j)
                            cluster.append(list(precedent_breakup.keys())[j])
                else:

                    dis_pet = lev(pet, pet_1)
                    dis_res = lev(res, res_1)

                    if dis_pet < threshold_pet and dis_res < threshold_res:
                        exclude.append(j)
                        cluster.append(list(precedent_breakup.keys())[j])

            precedent_clusters[cluster_num] = cluster
            cluster_num = cluster_num + 1

        elif cit != None:
            for j in range(i + 1, len(precedent_breakup)):
                cit_1 = list(precedent_breakup.values())[j][2]
                if cit_1 != None and cit_1 == cit:
                    exclude.append(j)
                    cluster.append(list(precedent_breakup.keys())[j])
            precedent_clusters[cluster_num] = cluster
            cluster_num = cluster_num + 1


    return precedent_clusters


def split_precedents(precedents):
    precedent_breakup = {}
    regex_vs = r'\b(?i)((v(\.|/)*s*\.*)|versus)\s+'
    regex_cit = r'(\(\d+\)|\d+|\[\d+\])\s*(\(\d+\)|\d+|\[\d+\])*\s*[A-Z\.]+\s*(\(\d+\)|\d+|\[\d+\])*\s*'

    for entity in precedents:
        citation = re.search(regex_cit, entity.text)
        if citation:
            cit = citation.group()
            text = entity.text[:citation.start()]
        else:
            cit = None
            text = entity.text
        vs = re.search(regex_vs, text)
        if vs:
            pet = (text[:vs.start()].strip())
            res = (text[vs.end():].strip())
            precedent_breakup[entity] = [pet, res, cit]
        else:

            precedent_breakup[entity] = [None, None, cit]

    return precedent_breakup


def merge_supras_precedents(precedent_supra_matches, precedent_clusters):
    counter = len(list(precedent_clusters.keys()))

    for i, s_p_match in enumerate(precedent_supra_matches.values()):
        c = 0
        for j, cluster in enumerate(precedent_clusters.values()):
            if s_p_match in cluster:
                c = 1
                cluster.append(list(precedent_supra_matches.keys())[i])
        if c == 0:
            precedent_clusters[counter] = [list(precedent_supra_matches.keys())[i], s_p_match]
            counter = counter + 1

    return precedent_clusters


def set_main_cluster(clusters):
    final_clusters = {}
    for c in clusters.keys():
        mains = max(clusters[c], key=len)
        final_clusters[mains] = clusters[c]
    return final_clusters


def precedent_coref_resol(doc):
    '''
  It is used for creating precedent clusters
    Parameters:
        doc : nlp doc object containing all the entity information
    Returns:
        clusters: dict with keys as head of each cluster and value is a list of all precedents in that cluster
   '''

    try:
        entities_noun = get_entities(doc, ['OTHER_PERSON', 'ORG', 'PETITIONER', 'RESPONDENT'])
        entities_precedents = get_entities(doc, ['PRECEDENT'])


        precedent_breakup = split_precedents(entities_precedents) ###split precedents into petitioner,respondent and citation

        precedent_clusters = create_precedent_clusters(precedent_breakup, percentage_threshold=20)  ###create precedent clusters based on fuzzy matching of petitioner,respondent and citation


        precedent_supra_matches, supras = get_precedent_supras(doc, entities_noun, precedent_breakup, entities_precedents,  ### get all the supras from judgment text and create entities
                                                               changes_threshold=2)

        precedent_supra_clusters = merge_supras_precedents(precedent_supra_matches, precedent_clusters)  ###asssign supras to the created clusters

        final_clusters = set_main_cluster(precedent_supra_clusters) ###assign head to each cluster
        clusters = {}
        entities = []

        for cluster in final_clusters.keys():
            if len(final_clusters[cluster]) > 1:
                clusters[cluster] = final_clusters[cluster]

        for entitiy in doc.ents:    ###change label of supras to precedents
            if entitiy in supras:
                entitiy.label_ = 'PRECEDENT'
                entities.append(entitiy)
            else:
                entities.append(entitiy)
        doc.ents = entities
    except:
        clusters = None
    return clusters


def get_roles(doc):
    other_person = []
    known_person = []
    entities = list(doc.ents)
    entities_to_remove = []

    for i, ents in enumerate(entities):
        if ents.label_ == 'OTHER_PERSON' or ents.label_=='ORG':

            entities_to_remove.append(ents)
            other_person.append(ents)
        elif ents.label_ == 'PETITIONER' or ents.label_ == 'RESPONDENT' or ents.label_ == 'JUDGE' or ents.label_ == 'WITNESS' or ents.label_ == 'LAWYER':
            known_person.append(ents)

    for ent in entities_to_remove:
        entities.remove(ent)

    return entities, other_person, known_person


def map_exact_other_person(doc):
    entities, other_person, known_person = get_roles(doc)

    other_person_text = [' '.join(oth.text.split()).lower().replace(',', '') for oth in other_person]

    ents_text = [' '.join(oth.text.split()).lower().replace(',', '') for oth in entities]
    count = 0
    other_person_found = []
    other_person_to_remove = []
    for i, other_p in enumerate(other_person):

        if other_person_text[i] in ents_text:
            labels = [entities[j].label_ for j, x in enumerate(ents_text) if other_person_text[i] == x]

            if len(set(labels)) == 1:
                count = count + 1
                other_person_to_remove.append(other_p)
                index = ents_text.index(other_person_text[i])

                other_person_found.append(other_p)
                if entities[index].label_ in ['PETITIONER', 'RESPONDENT', 'JUDGE', 'WITNESS', 'LAWYER']:
                    other_person_found[-1].label_ = entities[index].label_

    for oth in other_person_to_remove:
        other_person.remove(oth)

    return other_person, other_person_found, entities, known_person


def check_alias(names):
    names_text = [[' '.join(oth.text.split()).lower().replace(',', '').strip(), oth.label_] for oth in names]

    names_labels = []
    for i, name in enumerate(names_text):
        new_names = re.split('@|alias', name[0])
        if len(new_names) > 1:
            for n in new_names:
                names_labels.append([n.strip(), name[1], i])
        else:
            names_labels.append([name[0], name[1], i])

    return names_labels


def separate_name(names, only_first_last_name):
    aliased_cleaned_names = check_alias(names)
    separated_names = []
    for name in aliased_cleaned_names:
        separated = name[0].split(' ')
        if len(separated) > 1:
            if not only_first_last_name:
                separated_names.append([separated[-1], name[1], name[2]])
                separated_names.append([' '.join(separated[:-1]), name[1], name[2]])

        else:
            separated_names.append([separated[0], name[1], name[2]])

    return separated_names


def remove_ambiguous_names(known_person_cleaned):
    unique_known_person_cleaned = {}
    to_remove = []
    for i, el in enumerate(known_person_cleaned):

        if el[0] not in unique_known_person_cleaned.keys():
            unique_known_person_cleaned[el[0]] = [el[1]]
        else:
            unique_known_person_cleaned[el[0]].append(el[1])
    for kno in unique_known_person_cleaned.keys():
        if len(list(set(unique_known_person_cleaned[kno]))) > 1:
            to_remove.append(kno)
    known_person_left = []
    for kno in known_person_cleaned:
        if kno[0] not in to_remove:
            known_person_left.append(kno)
    known_person_cleaned_text = [other[0] for other in known_person_left]
    return known_person_cleaned_text, known_person_left


def map_name_wise_other_person(other_person_cleaned, known_person_cleaned):
    known_person_cleaned_text, known_person_left = remove_ambiguous_names(known_person_cleaned)
    c = 0
    other_person_found = []

    for i, other in enumerate(other_person_cleaned):

        if other[0] in known_person_cleaned_text:
            other_person_found.append([other[2], known_person_left[known_person_cleaned_text.index(other[0])][1]])

            c = c + 1
    return other_person_found


def other_person_coref_res(doc):
    try:
        other_person, other_person_found, entities, known_person = map_exact_other_person(doc)

        known_person_cleaned = separate_name(known_person, only_first_last_name=False)
        other_person_cleaned = separate_name(other_person, only_first_last_name=True)

        oth = map_name_wise_other_person(other_person_cleaned, known_person_cleaned)
        remove = []
        for o in oth:

            other_person[o[0]].label_ = o[1]
            if other_person[o[0]] not in other_person_found:
                other_person_found.append(other_person[o[0]])

        for person in other_person:

            if person.label_ == 'OTHER_PERSON' or person.label_=='ORG':
                other_person_found.append(person)

        other_person_found.extend(known_person)

    except:
        other_person_found = None


    return other_person_found


def remove_overlapping_entities(ents, pro_sta_clusters):
    try:
        final_ents = []
        for i in ents:
            if i.label_ not in ['PETITIONER', 'RESPONDENT', 'LAWYER', 'JUDGE', 'OTHER_PERSON', 'WITNESS', 'PROVISION','ORG']:
                final_ents.append(i)

        for cluster in pro_sta_clusters:
            if cluster[0] not in final_ents:
                final_ents.append(cluster[0])

        final_ents = spacy.util.filter_spans(final_ents)
    except:
        final_ents = None
    return final_ents


def get_exact_match_pro_statute(docs):
    pro_statute = []
    pro_left = []
    total_statutes = []
    total_pros = []

    for doc in docs.sents:

        statutes = []
        pros = []
        for ents in doc.ents:
            if ents.label_ == 'STATUTE':
                statutes.append(ents)
                total_statutes.append(ents)
            elif ents.label_ == 'PROVISION':
                pros.append(ents)
                total_pros.append(ents)

        for statute in statutes:

            start = statute.start
            nearest = []
            for pro in pros:
                if pro.end <= statute.start:
                    nearest.append(statute.start - pro.end)
            if len(nearest) > 0:
                provision_ind = nearest.index(min(nearest))
                provision = pros[provision_ind]
                pros.pop(provision_ind)
                pairs = [provision, statute]

                pro_statute.append(pairs)

        if len(pros) > 0:
            pro_left.extend(pros)

    return pro_statute, pro_left, total_statutes


def separate_provision_get_pairs_statute(pro_statute):
    matching_pro_statute = []
    to_remove = []
    sepearte_sec = r'(?i)(section(s)*|article(s)*)'
    remove_braces = r'\('
    sepearte_sub_sec = r'(?i)((sub|sub-)section(s)*|clause(s)*|annexure(s)*)'
    for pro in pro_statute:
        sub_section = re.split('of', pro[0].text)

        if len(sub_section) > 1:
            section = sub_section[1:]
        else:
            section = re.split(',|and|/|or', pro[0].text)

        for sec in section:
            match_sub_sec = re.search(sepearte_sub_sec, sec)
            if match_sub_sec:
                to_remove.append(pro)

                continue
            match_sec = re.search(sepearte_sec, sec)
            match_braces = re.search(remove_braces, sec)

            if match_braces:
                sec = sec[:match_braces.start()]

            if match_sec:
                sections = sec[match_sec.end():]

                matching_pro_statute.append([sections.strip(), pro[1]])

            else:

                matching_pro_statute.append([sec.strip(), pro[1]])

    return to_remove, matching_pro_statute


def check_validity(provision, statute):
    if 'article' in provision.text.lower():
        if 'constitution' in statute.text.lower() or 'rules' in statute.text.lower():
            return False
        else:
            return True

    else:
        if 'constitution' in statute.text.lower():
            return True
        else:
            return False


def map_pro_statute_on_heuristics(matching_pro_left, matching_pro_statute, pro_statute, total_statutes):
    co = 0
    if len(total_statutes) > 0:

        for pro_left in matching_pro_left:
            provision_to_find = pro_left[0]

            sta = [i for i, v in enumerate(matching_pro_statute) if v[0] == provision_to_find]
            j = 0
            for j, statute in enumerate(sta):
                if matching_pro_statute[statute][1].start > pro_left[1].end:
                    break

            if len(sta) > 0:

                if j > 0:
                    sta_index = j - 1
                else:
                    sta_index = 0
                statute = matching_pro_statute[sta[sta_index]]

                if pro_statute[-1][0] != pro_left[1]:

                    pro_statute.append([pro_left[1], statute[1]])

                    co = co + 1


                else:

                    pro_statute.pop(-1)
                    pro_statute.append([pro_left[1], statute[1]])




            else:

                i = 0
                for m, v in enumerate(total_statutes):

                    if v.end > pro_left[1].end:
                        i = m
                        break

                while check_validity(pro_left[1], total_statutes[i - 1]):
                    if i == 0:
                        break
                    i = i - 1

                if len(pro_statute) > 0:

                    if pro_statute[-1][0] != pro_left[1]:
                        matching_pro_statute.append([pro_left[0], total_statutes[i - 1]])

                        pro_statute.append([pro_left[1], total_statutes[i - 1], ''])

    return matching_pro_statute, pro_statute


def get_clusters(pro_statute):
    custom_ents = []
    k = 0
    clusters = []
    for pro in pro_statute:
        if len(pro) > 2:
            k = k + 1

            custom_ents.append(pro)
            pro.pop(2)
        else:
            clusters.append(pro)

    for ent in custom_ents:
        clusters.append((ent[0], ent[1]))

    return clusters


def separate_provision_get_pairs_pro(pro_left):
    matching_pro_left = []

    sepearte_sec = r'(?i)(section(s)*|article(s)*)'
    remove_braces = r'\('
    sepearte_sub_sec = r'(?i)(((sub|sub-)\s*section(s)*)|clause(s)*|annexure(s)*)'

    for pro in pro_left:
        sub_section = re.split('of', pro.text)
        if len(sub_section) > 1:
            section = sub_section[1:]
        else:
            section = re.split(',|and|/|or', pro.text)

        for sec in section:

            match_sub_sec = re.search(sepearte_sub_sec, sec)

            if match_sub_sec:
                continue
            match_sec = re.search(sepearte_sec, sec)
            match_braces = re.search(remove_braces, sec)

            if match_braces:
                sec = sec[:match_braces.start()]
            if len(sec.strip()) > 0:

                if match_sec:
                    sections = sec[match_sec.end():]
                    matching_pro_left.append([sections.strip(), pro])

                else:

                    matching_pro_left.append([sec.strip(), pro])
    return matching_pro_left


def pick_statute_from_multiple_statutes(statutes, threshold=5):
    levens = []
    indexes = {}
    exclude = []
    for i, statute in enumerate(statutes):
        if i in exclude:
            continue

        if len(statutes) - 1 > i:

            for j, stat in enumerate(statutes[i + 1:]):

                dist = lev(statute.text, stat.text)
                if dist < threshold:

                    if statute not in indexes.keys():
                        indexes[statute] = []

                    indexes[statute].append(statutes[i + 1 + j])

                    exclude.append(i + 1 + j)
        else:
            indexes[statute] = []

    max_value = (max(indexes.values(), key=len))

    if len(max_value) == 0:
        return list(indexes.keys())[0]
    else:
        return max_value[0]


def find_year_statute(statutes_left, total_statutes):
    ### to find statutes with only years eg. 1986 act
    regex_only_year = r'(?i)^\d{4}\s*act$'
    regex_find_year = r'.*([1-3][0-9]{3})'
    only_year_statutes = [statute for statute in statutes_left if re.findall(regex_only_year, statute.text)]
    statutes_with_years = {}
    statutes_to_find = []
    statutes_found = []
    for statute in total_statutes:
        if statute not in only_year_statutes:
            statute_year = []
            years = re.findall(regex_find_year, statute.text)
            if years:
                for year in years:
                    if year in statutes_with_years.keys():

                        statutes_with_years[year].append(statute)
                    else:
                        statutes_with_years[year] = []
                        statutes_with_years[year].append(statute)
        else:
            statutes_to_find.append(re.findall(regex_find_year, statute.text)[0])
    for i, statute in enumerate(statutes_to_find):
        if statute in statutes_with_years.keys():
            found = pick_statute_from_multiple_statutes(statutes_with_years[statute], threshold=5)
            statutes_found.append(found)
        else:
            statutes_found.append(only_year_statutes[i])

    # cluster=[]
    # for statute in statutes_found:
    #     if statute not in cluster.keys():
    #         cluster[statute]=[]
    #     cluster_statute.append()

    return only_year_statutes, statutes_found


def create_statute_clusters_using_lev(clusters, statutes, threshold=5):
    for i, statute in enumerate(statutes):
        done = 0
        for j, stat in enumerate(clusters.keys()):
            dist = lev(statute.text, stat)
            if dist < threshold:
                done = 1

                clusters[stat].append(statute)
        if done == 0:
            clusters[statute.text] = [statute]
    return clusters


def get_initials(statute):
    statute = re.sub(r"[\(\[].*?[\)\]]", "", statute.strip())
    statute_left = re.split(r'\s+', statute.strip().split(',')[0])  # remove year and split
    acronym = ''
    stat = ''

    for s in statute_left:

        if s.lower() not in ['act', 'rules', 'order', 'code', 'constitution', '']:

            acronym = acronym + s[0]
        else:
            stat = s

    return acronym, stat


def create_acronym(statute):
    acronym, stat = get_initials(statute)
    regex_act = r'(?i)({})'
    variation = ' '
    for a in acronym:
        if a[0].isalpha():
            if a[0].islower():
                variation = variation + a + r'*\.*\s*'
            else:
                variation = variation + a + r'\.*\s*'
    regex_act = regex_act.format(variation)
    regex_act = regex_act.replace(' ', '')
    if stat.strip() != '':
        if stat.lower() == 'code':
            regex_act = regex_act + stat[0] + r'*\.*[' + stat + ']*'  ##code can be written as c or code
        else:
            regex_act = regex_act + r'\.*(' + stat + r')\b'
    return regex_act


def provide_predefined_statute(fullform, acronyms):
    # if len(acronyms) > 0:
    acr = ''
    regex_act = r'(?i){}'
    regex_acronym = regex_act.format('(\\b' + '\\b|\\b'.join(acronyms) + '\\b)')
    # else:
    #     regex_acronym = create_acronym(fullform)
    return regex_acronym


def remove_year(statute):
    year = re.findall(r'.*([1-3][0-9]{3})', statute)
    for y in year:
        statute = statute.replace(y, '')
    statute = statute.replace(',', '').strip()

    return statute


def find_acronym_statute(statutes_left, total_statutes, acr_dict):
    regex_check_acronym = r"([A-Z]+[a-z]{0,1}\.*\s*,*)*((A|a)(c|C)(t|T))*\s*"
    to_find = []
    to_find_statute = []
    found = []
    acronym_ff_pairs = {}
    for statute in total_statutes:

        if statute in statutes_left and re.fullmatch(regex_check_acronym, remove_year(statute.text)):
            to_find.append(remove_year(statute.text))
            to_find_statute.append(statute)
        else:
            acr = create_acronym(statute.text)
            if acr not in acronym_ff_pairs.keys():
                acronym_ff_pairs[acr] = []
            acronym_ff_pairs[acr].append(statute)

    for k, statute in enumerate(to_find):
        counter = 0

        for acr in acronym_ff_pairs.keys():
            if re.fullmatch(acr, statute):
                found.append(acronym_ff_pairs[acr][0])
                counter = 1
                break

        if counter == 0:
            found.append(to_find_statute[k])
    return to_find_statute, found


def merge_clusters(clusters, threshold=5):
    levens = []
    new_clusters = {}
    exclude = []

    statutes = list(clusters.keys())
    for i, statute in enumerate(statutes):
        done = 0
        if i in exclude:
            continue

        if len(statutes) - 1 > i:

            for j, stat in enumerate(statutes[i + 1:]):

                dist = lev(statute, stat)
                if dist < threshold:
                    done = 1

                    if statute not in new_clusters.keys():
                        new_clusters[statute] = clusters[list(clusters.keys())[i]]

                    new_clusters[statute].extend(clusters[list(clusters.keys())[i + 1 + j]])

                    exclude.append(i + 1 + j)
            if done == 0:
                new_clusters[statute] = clusters[list(clusters.keys())[i]]


        else:
            new_clusters[statute] = clusters[list(clusters.keys())[i]]
    return new_clusters


def statute_clusters_with_years(statutes):
        clusters={}

        done=[]


        regex_find_year = r'.*([1-3][0-9]{3})'
        statute_cleaned=[]
        years_cleaned=[]
        for statute in statutes:
                    years = re.findall(regex_find_year, statute.text)
                    year=''
                    if len(re.findall(regex_find_year, statute.text))>0:

                        year=years[0]

                    statute_cleaned.append(statute.text.replace(year,'').replace(',','').strip().lower())
                    years_cleaned.append(year)


        indices_done=[]
        for m,statute in enumerate(statute_cleaned):

            if m  in indices_done:
                continue
            indices = [i for i, x in enumerate(statute_cleaned) if x == statute]
            indices_done.extend(indices)
            no_years=[]
            with_years=[]


            if len(indices)>1:
                for indice  in indices:
                    if years_cleaned[indice]=='':
                        no_years.append(indice)
                    else:
                        with_years.append(int(years_cleaned[indice]))
                if len(with_years)>1:
                    max_indice=years_cleaned.index(str((max(with_years))))
                    clusters[statutes[max_indice].text] = [statutes[max_indice]]
                    done.append(statutes[max_indice])
                    for ind in no_years:
                        clusters[statutes[max_indice].text].append(statutes[ind])
                        done.append(statutes[ind])



                else:
                    if len(with_years)==1:
                        indi=years_cleaned.index(str(with_years[0]))
                    else:
                        if len(no_years)>0:
                            indi=years_cleaned.index('')
                        else:
                            continue
                    clusters[statutes[indi].text]=[]
                    for ind in indices:
                        clusters[statutes[indi].text].append(statutes[ind])
                        done.append(statutes[ind])



        not_done=[statute for statute in statutes if statute not in done]

        return clusters,not_done








def create_statute_clusters(doc, old_statute_clusters, new_statute_clusters, statute_shortforms_path):
    clusters = {}
    statutes = []
    not_done = []
    acr_dict = {}
    if statute_shortforms_path != '':
        statute_shortforms = pd.read_csv(statute_shortforms_path)
        fullforms = list(statute_shortforms['fullforms'])

        sf = list(statute_shortforms['shortforms'])

        for i, f in enumerate(fullforms):
            if f not in acr_dict:
                acr_dict[f] = []
            acr_dict[f].append(sf[i])

    for ent in doc.ents:
        if ent.label_ == 'STATUTE':
            statutes.append(ent)

    for c in old_statute_clusters.keys():
        if c not in clusters.keys():
            clusters[c.text] = old_statute_clusters[c]
        else:
            clusters[c.text].extend(old_statute_clusters[c])

    for c in new_statute_clusters.keys():
        if c not in clusters.keys():
            clusters[c.text] = new_statute_clusters[c]
        else:
            clusters[c.text].extend(new_statute_clusters[c])

    flat_list = [item for sublist in clusters.keys() for item in clusters[sublist]]
    not_done = [item for item in statutes if item not in flat_list]

    for i,statute in enumerate(not_done):

        stat = check_stat(statute.text, acr_dict)
        if stat == '':
            # not_done.append(statute)
            continue
        if stat in clusters.keys():
            clusters[stat].append(statute)
        else:
            clusters[stat] = []
            clusters[stat].append(statute)
    flat_list = [item for sublist in clusters.keys() for item in clusters[sublist]]
    not_done = [item for item in statutes if item not in flat_list]

    new_clusters,not_done=statute_clusters_with_years(not_done)

    clusters.update(new_clusters)
    statutes_to_find_year, statutes_found_year = find_year_statute(not_done, statutes)
    statutes_left = [stat for stat in not_done if stat not in statutes_to_find_year]
    total_statutes_left = [stat for stat in statutes if stat not in statutes_to_find_year]

    statutes_to_find_acronym, statutes_found_acroym = find_acronym_statute(statutes_left, total_statutes_left, acr_dict)
    statutes_to_find_year.extend(statutes_to_find_acronym)
    statutes_found_year.extend(statutes_found_acroym)

    for j, statute in enumerate(statutes_found_year):

        checker = 0
        for cluster in clusters.keys():
            if statute in clusters[cluster]:
                clusters[cluster].append(statutes_to_find_year[j])
                checker = checker + 1
                break
        if checker == 0:
            if statute.text not in clusters.keys():
                clusters[statute.text] = []
            clusters[statute.text].append(statutes_to_find_year[j])
        # else:
        #     clusters[statute.text]=[]
        #     clusters[statute.text].append(statutes_to_find_year[j])

    flat_list = [item for sublist in clusters.keys() for item in clusters[sublist]]
    not_done = [item for item in statutes if item not in flat_list]
    new_clusters = merge_clusters(clusters)
    final_clusters = create_statute_clusters_using_lev(new_clusters, not_done, threshold=5)

    # (item for sublist in clusters_lev for item in sublist)
    # for statutes in clusters_lev:
    #     for j,statute in enumerate(statutes_found_year):
    #         if statute in statutes:
    #             statutes.append(statutes_found_year[j])

    return final_clusters


def check_stat(text, acr_dict):
    regex_crpc = r'(?i)\b(((criminal|cr)\.*\s*(procedure|p)\.*\s*(c|code)\.*)|(code\s*of\s*criminal\s*procedure))\s*'
    regex_ipc = r'(?i)\b((i|indian)+\.*\s*(penal|p)\.*\s*(c|code))\.*'
    regex_cons = r'(?i)\b((constitution)+\s*(of\s*india\s*)*)\b'
    regex_itact = r'(?i)\b((i\.*\s*t\.*\s*|income\s*\-*tax\s+)act\s*)\b'
    regex_mvact = r'(?i)\b((m\.*\s*v\.*\s*)|(motor\s*\-*vehicle(s)*\s+)act\s*)\b'
    regex_idact = r'(?i)\b((i\.*\s*d\.*\s*)|(industrial\s*\-*dispute(s)*\s+)act\s*)\b'
    regex_sarfaesi = r'(?i)\b((s\.*\s*a\.*\s*r\.*\s*f\.*\s*a\.*\s*e\.*\s*s\.*\s*i\.*\s*)|(securitisation\s*and\s*reconstruction\s*of\s*financial\s*assets\s*and\s*enforcement\s*of\s*security\s*interest\s+)act\s*)\b'
    regex_cpc = r'(?i)\b(((civil|c)\.*\s*(procedure|p)\.*\s*(c|code)\.*)|(code\s*of\s*civil\s*procedure))\s*'
    regex_ndpc = r'(?i)\b((narcotic|n\.*)\s*(drugs\s*and|d\.*)\s*(psychotropic|p\.*)\s*(substances|s\.*)\s*(act)*)\s*'
    regex_ni = r'(?i)\b((negotiable|n\.*)\s*(instruments|i\.*)\s*(act))\s*'
    regex_pocso = r'(?i)\b((prevention|p\.*)\s*(of|o\.*)\s*(children|c\.*)\s*(from)*\s*(sexual|s\.*)\s*(offences|o\.*)\s*(act)*\s*)'
    regex_pota = r'(?i)\b((prevention|p\.*)\s*(of|o\.*)\s*(terrorism|t\.*)\s*(act|a\.*)\s*)'
    regex_tada = r'(?i)\b((terrorist|t\.*)\s*(and|a\.*)\s*(disruptive|d\.*)\s*(activities|a\.*)\s*)'
    regex_tp = r'(?i)\b((transfer|t\.*)\s*(of)*\s*(property|p\.*)\s*(act))\s*'

    match_crpc = re.search(regex_crpc, text)
    match_ipc = re.search(regex_ipc, text)
    match_cons = re.search(regex_cons, text)
    match_ita = re.search(regex_itact, text)
    match_mv = re.search(regex_mvact, text)
    match_idact = re.search(regex_idact, text)
    match_sarfaesi = re.search(regex_sarfaesi, text)
    match_cpc = re.search(regex_cpc, text)

    match_ndpc = re.search(regex_ndpc, text)
    match_ni = re.search(regex_ni, text)
    match_pocso = re.search(regex_pocso, text)
    match_pota = re.search(regex_pota, text)
    match_tada = re.search(regex_tada, text)
    match_tp = re.search(regex_tp, text)
    if len(acr_dict) > 0:

        for fullform in acr_dict.keys():
            regex = provide_predefined_statute(fullform, acr_dict[fullform])

            if re.search(regex, text):
                return fullform
    if match_crpc:
        return 'Code of Criminal Procedure '
    elif match_ipc:
        return 'Indian Penal Code'
    elif match_cons:
        return 'Constitution of India'
    elif match_ita:
        return 'Income Tax Act'
    elif match_mv:
        return 'Motor Vehicle Act'
    elif match_idact:
        return 'Industrial Dispute Act'
    elif match_sarfaesi:
        return 'Securitisation and Reconstruction of Financial Assets and Enforcement of Securities Interest Act'
    elif match_cpc:
        return 'Code of Civil Procedure'
    elif text.lower().strip() == 'sebi act':
        return 'Securities and Exchange Board of India Act'
    elif match_ndpc:
        return 'Narcotic Drugs and Psychotropic Substances Act'
    elif match_ni:

        return 'Negotiable Instruments act'
    elif match_pocso:
        return 'Protection Of Children from Sexual Offences Act'
    elif match_pota:
        return 'Prevention of Terrorism Act'
    elif match_tada:
        return 'Terrorist and Disruptive Activities Act'
    elif match_tp:
        return 'Transfer of Property Act'

    else:
        return ''


def remove_unidentified_statutes(doc, new_statutes):
    entities = doc.ents
    stats = []
    new_entities = []

    stats.extend(new_statutes)

    for ents in entities:
        if ents not in stats:
            new_entities.append(ents)

    return new_entities


def create_unidentified_statutes(doc):
    # regex=r'(?i)\((\s*.*\s*act.*\)?)'
    regex = r'\((.*?)\)'
    clusters_new_statutes = {}
    statutes = []
    for ent in doc.ents:
        if ent.label_ == 'STATUTE':
            statutes.append(ent)

    statutes_start_end = [(sta.start, sta.end) for sta in statutes]
    statutes_text = [statute.text for statute in statutes]
    for statute in statutes:
        end_char = statute.end_char
        text = doc.text[end_char:]
        match = re.search(regex, text)

        if match and match.span()[0] == 1:

            # regex_act=r'\b(?i).*act\s*'
            regex_act = r"\b(([A-Z][A-Za-z'']*|\d{4})(?:\s+[A-Z][a-z'']*)*\s*(a|A)ct|\s*(a|A)ct)\b"  ###to match consecutive  words starting with upper case or years followed by the word act

            match1 = re.search(regex_act, match.group())

            if match1:

                stat_text = match1.group()

                if statute not in clusters_new_statutes.keys():
                    clusters_new_statutes[statute] = []
                    clusters_new_statutes[statute].append(stat_text.strip())
                else:
                    clusters_new_statutes[statute].append(stat_text.strip())

    new_statutes = []
    new_statutes_clusters = {}
    text = doc.text
    for statute in clusters_new_statutes.keys():
        for sta in clusters_new_statutes[statute]:
            ent = re.finditer(sta, text)

            stat_new = [doc.char_span(e.start(), e.end(), label="STATUTE", alignment_mode='expand') for e in ent]
            new_statutes.extend(stat_new)
            if sta not in new_statutes_clusters.keys():
                new_statutes_clusters[statute] = []
                new_statutes_clusters[statute].extend(stat_new)
            else:
                new_statutes_clusters[statute].extend(stat_new)

    discarded_statutes = []
    for sta in new_statutes:
        for s in statutes_start_end:
            if sta.start >= s[0] and sta.end <= s[1]:
                discarded_statutes.append(sta)
    old_statute_clusters = {}

    for s in discarded_statutes:
        if s in new_statutes:
            new_statutes.remove(s)

    for sta in new_statutes_clusters.keys():
        for s in new_statutes_clusters[sta]:

            if s in discarded_statutes:

                new_statutes_clusters[sta].remove(s)

                if sta in old_statute_clusters.keys():
                    old_statute_clusters[sta].append(s)
                else:
                    old_statute_clusters[sta] = []
                    old_statute_clusters[sta].append(s)

    return new_statutes_clusters, new_statutes, old_statute_clusters


def add_statute_head(clusters, stat_clusters):
    new_clusters = []
    clusters_done = []
    provision_statutes = collections.namedtuple('provision_statutes',
                                                ['provision_entity', 'statute_entity', 'normalised_provision_text',
                                                 'normalised_statute_text'])

    for stat_cluster in stat_clusters.keys():
        acts = stat_clusters[stat_cluster]

        for i, cluster in enumerate(clusters):
            if cluster[1] in acts:
                new_clusters.append(provision_statutes(cluster[0], cluster[1], cluster[2], stat_cluster))

                clusters_done.append(cluster)

    k = 0
    for cluster in clusters:
        if cluster not in clusters_done:
            k = k + 1
            new_clusters.append(provision_statutes(cluster[0], cluster[1], cluster[2], cluster[1].text))

    return new_clusters


def pro_statute_coref_resol(doc, statute_shortforms_path):
    # clusters_lev=create_statute_clusters_using_lev(statutes,threshold=5)
    '''
     It is used for creating statute and provision clusters
       Parameters:
           doc : nlp doc object containing all the entity information
           statute_shortforms_path: path to the csv containing predefined statute shortforms
       Returns:
           provision_clusters:  list of named tuples, each tuple contains provision-statute-normalised provision-normalised statutes text
           stat_clusters:  dict with keys as head of each cluster and value is a list of all statutes in that cluster
      '''
    try:


        new_statutes_clusters, new_statutes, old_statute_clusters = create_unidentified_statutes(doc)  ###create statutes that are within the parantheses like hereby referred as 'act'
        old_entities = list(doc.ents)

        for ent in new_statutes:
            if ent not in old_entities:
                old_entities.append(ent)
        old_entities = spacy.util.filter_spans(old_entities)

        doc.ents = old_entities

        stat_clusters = create_statute_clusters(doc, old_statute_clusters, new_statutes_clusters,
                                                statute_shortforms_path)  ###create cluster of statutes based on matching exact statute names,by years and by acronyms
        pro_statute, pro_left, total_statutes = get_exact_match_pro_statute(doc)  ###create pairs of explicit provision-statutes
        to_remove, matching_pro_statute = separate_provision_get_pairs_statute(pro_statute)  ###separate provisions like section 2 and 3 and remove only subsection and clauses
        matching_pro_left = separate_provision_get_pairs_pro(pro_left)  ###separate provisions for the left provisions

        for pro in to_remove:
            if pro in pro_statute:
                pro_statute.remove(pro)

        matching_pro_statute, pro_statute = map_pro_statute_on_heuristics(matching_pro_left,
                                                                          matching_pro_statute,
                                                                          pro_statute,
                                                                          total_statutes)   ###map left provisions to statutes based on the heuristics

        clusters = get_clusters(pro_statute)  ###assign each provision to the statute cluster it belongs to

        clusters = seperate_provision(doc, clusters)  ###separate provisions and create new entities for them

        new_entities = remove_unidentified_statutes(doc, new_statutes)  ###remove the extra statutes identified initially
        doc.ents = new_entities

        # for cluster in new_statutes_clusters.keys():
        #     stat_clusters[cluster.text] = new_statutes_clusters[cluster]

        provision_clusters = add_statute_head(clusters, stat_clusters)   ###add statute head to each provision-statute pair
    except:
        provision_clusters, stat_clusters = None, None

    return provision_clusters, stat_clusters


def seperate_provision(doc, clusters):
    new_clusters = []

    for cluster in clusters:

        provision = cluster[0]
        statute = cluster[1]
        section = re.split(',| and |/| or |&', provision.text)
        start = provision.start_char
        pro = provision.text
        keyword = section[0].split(' ')[0].strip()
        if len(keyword) > 0:

            if keyword[-1] == 's':
                keyword = keyword[:-1]

        combined = False
        for sec in section:
            sec_text = sec.strip()
            if len(sec_text) > 0:
                if sec_text.replace(' ', '').isalpha() or (not sec_text[0].isnumeric() and not sec_text[0].isalpha()):
                    combined = True
                    break

        if len(section) > 1 and not combined:
            for sec in section:

                ind = pro.find(sec)
                sect = doc.char_span(start + ind, start + ind + len(sec), "PROVISION", alignment_mode='expand')
                pro = pro[ind + len(sec):]
                start = start + ind + len(sec)
                if len(sec.strip()) > 0:
                    if not sec.strip()[0].isalpha():
                        new_clusters.append((sect, statute, keyword + ' ' + sect.text))
                    else:
                        new_clusters.append((sect, statute, keyword + ' ' + ' '.join(sect.text.split(' ')[1:])))

        else:
            new_clusters.append((cluster[0], cluster[1], cluster[0].text))

    return new_clusters



"""
Implementation of "Attention is All You Need"
"""

import numpy as np
import torch
import torch.nn as nn

from .encoder import PositionalEncoding
from .neural import MultiHeadedAttention, PositionwiseFeedForward, DecoderState

MAX_SIZE = 5000


class TransformerDecoderLayer(nn.Module):
    """
    Args:
      d_model (int): the dimension of keys/values/queries in
                       MultiHeadedAttention, also the input size of
                       the first-layer of the PositionwiseFeedForward.
      heads (int): the number of heads for MultiHeadedAttention.
      d_ff (int): the second-layer of the PositionwiseFeedForward.
      dropout (float): dropout probability(0-1.0).
      self_attn_type (string): type of self-attention scaled-dot, average
    """

    def __init__(self, d_model, heads, d_ff, dropout):
        super(TransformerDecoderLayer, self).__init__()

        self.self_attn = MultiHeadedAttention(
            heads, d_model, dropout=dropout)

        self.context_attn = MultiHeadedAttention(
            heads, d_model, dropout=dropout)
        self.feed_forward = PositionwiseFeedForward(d_model, d_ff, dropout)
        self.layer_norm_1 = nn.LayerNorm(d_model, eps=1e-6)
        self.layer_norm_2 = nn.LayerNorm(d_model, eps=1e-6)
        self.drop = nn.Dropout(dropout)
        mask = self._get_attn_subsequent_mask(MAX_SIZE)
        # Register self.mask as a buffer in TransformerDecoderLayer, so
        # it gets TransformerDecoderLayer's cuda behavior automatically.
        self.register_buffer('mask', mask)

    def forward(self, inputs, memory_bank, src_pad_mask, tgt_pad_mask,
                previous_input=None, layer_cache=None, step=None):
        """
        Args:
            inputs (`FloatTensor`): `[batch_size x 1 x model_dim]`
            memory_bank (`FloatTensor`): `[batch_size x src_len x model_dim]`
            src_pad_mask (`LongTensor`): `[batch_size x 1 x src_len]`
            tgt_pad_mask (`LongTensor`): `[batch_size x 1 x 1]`

        Returns:
            (`FloatTensor`, `FloatTensor`, `FloatTensor`):

            * output `[batch_size x 1 x model_dim]`
            * attn `[batch_size x 1 x src_len]`
            * all_input `[batch_size x current_step x model_dim]`

        """
        dec_mask = torch.gt(tgt_pad_mask +
                            self.mask[:, :tgt_pad_mask.size(1),
                            :tgt_pad_mask.size(1)], 0)
        input_norm = self.layer_norm_1(inputs)
        all_input = input_norm
        if previous_input is not None:
            all_input = torch.cat((previous_input, input_norm), dim=1)
            dec_mask = None

        query = self.self_attn(all_input, all_input, input_norm,
                               mask=dec_mask,
                               layer_cache=layer_cache,
                               type="self")

        query = self.drop(query) + inputs

        query_norm = self.layer_norm_2(query)
        mid = self.context_attn(memory_bank, memory_bank, query_norm,
                                mask=src_pad_mask,
                                layer_cache=layer_cache,
                                type="context")
        output = self.feed_forward(self.drop(mid) + query)

        return output, all_input
        # return output

    def _get_attn_subsequent_mask(self, size):
        """
        Get an attention mask to avoid using the subsequent info.

        Args:
            size: int

        Returns:
            (`LongTensor`):

            * subsequent_mask `[1 x size x size]`
        """
        attn_shape = (1, size, size)
        subsequent_mask = np.triu(np.ones(attn_shape), k=1).astype('uint8')
        subsequent_mask = torch.from_numpy(subsequent_mask)
        return subsequent_mask


class TransformerDecoder(nn.Module):
    """
    The Transformer decoder from "Attention is All You Need".


    .. mermaid::

       graph BT
          A[input]
          B[multi-head self-attn]
          BB[multi-head src-attn]
          C[feed forward]
          O[output]
          A --> B
          B --> BB
          BB --> C
          C --> O


    Args:
       num_layers (int): number of encoder layers.
       d_model (int): size of the model
       heads (int): number of heads
       d_ff (int): size of the inner FF layer
       dropout (float): dropout parameters
       embeddings (:obj:`onmt.modules.Embeddings`):
          embeddings to use, should have positional encodings
       attn_type (str): if using a seperate copy attention
    """

    def __init__(self, num_layers, d_model, heads, d_ff, dropout, embeddings):
        super(TransformerDecoder, self).__init__()

        # Basic attributes.
        self.decoder_type = 'transformer'
        self.num_layers = num_layers
        self.embeddings = embeddings
        self.pos_emb = PositionalEncoding(dropout, self.embeddings.embedding_dim)

        # Build TransformerDecoder.
        self.transformer_layers = nn.ModuleList(
            [TransformerDecoderLayer(d_model, heads, d_ff, dropout)
             for _ in range(num_layers)])

        self.layer_norm = nn.LayerNorm(d_model, eps=1e-6)

    def forward(self, tgt, memory_bank, state, memory_lengths=None,
                step=None, cache=None, memory_masks=None):
        """
        See :obj:`onmt.modules.RNNDecoderBase.forward()`
        """

        src_words = state.src
        tgt_words = tgt
        src_batch, src_len = src_words.size()
        tgt_batch, tgt_len = tgt_words.size()

        # Run the forward pass of the TransformerDecoder.
        # emb = self.embeddings(tgt, step=step)
        emb = self.embeddings(tgt)
        assert emb.dim() == 3  # len x batch x embedding_dim

        output = self.pos_emb(emb, step)

        src_memory_bank = memory_bank
        padding_idx = self.embeddings.padding_idx
        tgt_pad_mask = tgt_words.data.eq(padding_idx).unsqueeze(1) \
            .expand(tgt_batch, tgt_len, tgt_len)

        if (not memory_masks is None):
            src_len = memory_masks.size(-1)
            src_pad_mask = memory_masks.expand(src_batch, tgt_len, src_len)

        else:
            src_pad_mask = src_words.data.eq(padding_idx).unsqueeze(1) \
                .expand(src_batch, tgt_len, src_len)

        if state.cache is None:
            saved_inputs = []

        for i in range(self.num_layers):
            prev_layer_input = None
            if state.cache is None:
                if state.previous_input is not None:
                    prev_layer_input = state.previous_layer_inputs[i]
            output, all_input \
                = self.transformer_layers[i](
                output, src_memory_bank,
                src_pad_mask, tgt_pad_mask,
                previous_input=prev_layer_input,
                layer_cache=state.cache["layer_{}".format(i)]
                if state.cache is not None else None,
                step=step)
            if state.cache is None:
                saved_inputs.append(all_input)

        if state.cache is None:
            saved_inputs = torch.stack(saved_inputs)

        output = self.layer_norm(output)

        # Process the result and update the attentions.

        if state.cache is None:
            state = state.update_state(tgt, saved_inputs)

        return output, state

    def init_decoder_state(self, src, memory_bank,
                           with_cache=False):
        """ Init decoder state """
        state = TransformerDecoderState(src)
        if with_cache:
            state._init_cache(memory_bank, self.num_layers)
        return state


class TransformerDecoderState(DecoderState):
    """ Transformer Decoder state base class """

    def __init__(self, src):
        """
        Args:
            src (FloatTensor): a sequence of source words tensors
                    with optional feature tensors, of size (len x batch).
        """
        self.src = src
        self.previous_input = None
        self.previous_layer_inputs = None
        self.cache = None

    @property
    def _all(self):
        """
        Contains attributes that need to be updated in self.beam_update().
        """
        if (self.previous_input is not None
                and self.previous_layer_inputs is not None):
            return (self.previous_input,
                    self.previous_layer_inputs,
                    self.src)
        else:
            return (self.src,)

    def detach(self):
        if self.previous_input is not None:
            self.previous_input = self.previous_input.detach()
        if self.previous_layer_inputs is not None:
            self.previous_layer_inputs = self.previous_layer_inputs.detach()
        self.src = self.src.detach()

    def update_state(self, new_input, previous_layer_inputs):
        state = TransformerDecoderState(self.src)
        state.previous_input = new_input
        state.previous_layer_inputs = previous_layer_inputs
        return state

    def _init_cache(self, memory_bank, num_layers):
        self.cache = {}

        for l in range(num_layers):
            layer_cache = {
                "memory_keys": None,
                "memory_values": None
            }
            layer_cache["self_keys"] = None
            layer_cache["self_values"] = None
            self.cache["layer_{}".format(l)] = layer_cache

    def repeat_beam_size_times(self, beam_size):
        """ Repeat beam_size times along batch dimension. """
        self.src = self.src.data.repeat(1, beam_size, 1)

    def map_batch_fn(self, fn):
        def _recursive_map(struct, batch_dim=0):
            for k, v in struct.items():
                if v is not None:
                    if isinstance(v, dict):
                        _recursive_map(v)
                    else:
                        struct[k] = fn(v, batch_dim)

        self.src = fn(self.src, 0)
        if self.cache is not None:
            _recursive_map(self.cache)



"""
An `Activation` is just a function
that takes some parameters and returns an element-wise activation function.
For the most part we just use
[PyTorch activations](https://pytorch.org/docs/master/nn.html#non-linear-activations).
Here we provide a thin wrapper to allow registering them and instantiating them `from_params`.

The available activation functions include

* "linear"
* ["mish"](https://arxiv.org/abs/1908.08681)
* ["swish"](https://arxiv.org/abs/1710.05941)
* ["relu"](https://pytorch.org/docs/master/nn.html#torch.nn.ReLU)
* ["relu6"](https://pytorch.org/docs/master/nn.html#torch.nn.ReLU6)
* ["elu"](https://pytorch.org/docs/master/nn.html#torch.nn.ELU)
* ["prelu"](https://pytorch.org/docs/master/nn.html#torch.nn.PReLU)
* ["leaky_relu"](https://pytorch.org/docs/master/nn.html#torch.nn.LeakyReLU)
* ["threshold"](https://pytorch.org/docs/master/nn.html#torch.nn.Threshold)
* ["hardtanh"](https://pytorch.org/docs/master/nn.html#torch.nn.Hardtanh)
* ["sigmoid"](https://pytorch.org/docs/master/nn.html#torch.nn.Sigmoid)
* ["tanh"](https://pytorch.org/docs/master/nn.html#torch.nn.Tanh)
* ["log_sigmoid"](https://pytorch.org/docs/master/nn.html#torch.nn.LogSigmoid)
* ["softplus"](https://pytorch.org/docs/master/nn.html#torch.nn.Softplus)
* ["softshrink"](https://pytorch.org/docs/master/nn.html#torch.nn.Softshrink)
* ["softsign"](https://pytorch.org/docs/master/nn.html#torch.nn.Softsign)
* ["tanhshrink"](https://pytorch.org/docs/master/nn.html#torch.nn.Tanhshrink)
* ["selu"](https://pytorch.org/docs/master/nn.html#torch.nn.SELU)
"""

import math

import torch

from opennyai.rhetorical_roles.allennlp_helper.common.registrable import Registrable


class Activation(torch.nn.Module, Registrable):
    """
    Pytorch has a number of built-in activation functions.  We group those here under a common
    type, just to make it easier to configure and instantiate them `from_params` using
    `Registrable`.

    Note that we're only including element-wise activation functions in this list.  You really need
    to think about masking when you do a softmax or other similar activation function, so it
    requires a different API.
    """

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError


# There are no classes to decorate, so we hack these into Registrable._registry.
# If you want to instantiate it, you can do like this:
# Activation.by_name('relu')()
Registrable._registry[Activation] = {
    "relu": (torch.nn.ReLU, None),
    "relu6": (torch.nn.ReLU6, None),
    "elu": (torch.nn.ELU, None),
    "gelu": (torch.nn.GELU, None),
    "prelu": (torch.nn.PReLU, None),
    "leaky_relu": (torch.nn.LeakyReLU, None),
    "threshold": (torch.nn.Threshold, None),
    "hardtanh": (torch.nn.Hardtanh, None),
    "sigmoid": (torch.nn.Sigmoid, None),
    "tanh": (torch.nn.Tanh, None),
    "log_sigmoid": (torch.nn.LogSigmoid, None),
    "softplus": (torch.nn.Softplus, None),
    "softshrink": (torch.nn.Softshrink, None),
    "softsign": (torch.nn.Softsign, None),
    "tanhshrink": (torch.nn.Tanhshrink, None),
    "selu": (torch.nn.SELU, None),
}


@Activation.register("linear")
class LinearActivation(Activation):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x


@Activation.register("mish")
class MishActivation(Activation):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * torch.tanh(torch.nn.functional.softplus(x))


@Activation.register("swish")
class SwishActivation(Activation):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x * torch.sigmoid(x)


@Activation.register("gelu_new")
class GeluNew(Activation):
    """
    Implementation of the GELU activation function currently in Google BERT repo (identical to OpenAI GPT). Also
    see the Gaussian Error Linear Units paper: https://arxiv.org/abs/1606.08415
    """

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return (
                0.5
                * x
                * (1.0 + torch.tanh(math.sqrt(2.0 / math.pi) * (x + 0.044715 * torch.pow(x, 3.0))))
        )


@Activation.register("gelu_fast")
class GeluFast(Activation):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return 0.5 * x * (1.0 + torch.tanh(x * 0.7978845608 * (1.0 + 0.044715 * x * x)))



class Record:
    def __init__(self, length, data):
        self.length = length
        self.data = data

    def to_tensor(self, pad_len):
        return self.data.to_tensor(pad_len)

    def get_data_length(self):
        return self.length

    def __str__(self):
        return "l:" + str(self.length)

    def __repr__(self):
        return "l:" + str(self.length)


class BucketBatch:

    @staticmethod
    def wrap_record(record):
        b = BucketBatch(max_record_len=record.get_data_length(), batch_size=1)
        b.add_record(record)
        return b

    def __init__(self, max_record_len, batch_size):
        self.max_record_len = max_record_len
        self.batch_size = batch_size
        self.records = []

    def get_len(self):
        return len(self.records)

    def to_tensor(self, record_data_to_tensor_func, merge_records_func):
        result = None
        for r in self.records:
            r = record_data_to_tensor_func(r, self.max_record_len)
            result = merge_records_func(result, r)
        return result

    def is_full(self):
        '''Batch is full if no records cannot be inserted anymore without violating batch_size'''
        return (len(self.records) + 1) * self.max_record_len > self.batch_size

    def is_empty(self):
        return len(self.records) == 0

    def can_add_record(self, record):
        return not self.is_full() and record.length <= self.max_record_len

    def add_record(self, record):
        self.records.append(record)

    def pop_record(self):
        r = self.records.pop(0)
        return r

    def peek_record(self):
        return self.records[0]

    def __str__(self):
        return "{max_record_len:" + str(self.max_record_len) + ", length:" + str(
            len(self.records)) + ", records=" + str(self.records) + "}"

    def __repr__(self):
        return self.__str__()

    def take_records_from_bucket(self, bucket):
        while not self.is_full() and not bucket.is_empty():
            if self.can_add_record(bucket.peek_record()):
                self.add_record(bucket.pop_record())
            else:
                break


def bucket_records(records, batch_size):
    len_to_buckets = dict()

    # group records with same length to buckets
    for r in records:
        record_len = r.length
        if record_len not in len_to_buckets:
            len_to_buckets[record_len] = [BucketBatch(record_len, batch_size)]

        bucket = len_to_buckets[record_len][-1]
        if not bucket.can_add_record(r):
            bucket = BucketBatch(record_len, batch_size)
            len_to_buckets[record_len].append(bucket)

        bucket.add_record(r)

        # Flatten buckets into a list
    buckets_list = [b for sublist in len_to_buckets.values() for b in sublist]

    # sort list of buckets: first come buckets with largest records 
    buckets_list = sorted(buckets_list, key=lambda b: b.max_record_len, reverse=True)

    # Merge buckets: move records of buckets with smaller records to buckets with larger records
    for i in range(0, len(buckets_list)):
        b = buckets_list[i]

        if b.is_full() or b.is_empty():
            continue

        for j in range(i + 1, len(buckets_list)):
            nb = buckets_list[j]

            if nb.is_empty():
                continue

            b.take_records_from_bucket(nb)

            if b.is_full():
                break

                # filter empty buckets
    buckets_list = [b for b in buckets_list if not b.is_empty()]

    return buckets_list



import re
import warnings

import spacy


def remove_unwanted_text(text):
    '''Looks for pattern  which typically starts the main text of jugement.
    The text before this pattern contains metadata like name of paries, judges and hence removed'''
    pos_list = []
    len = 0
    pos = 0
    pos_list.append(text.find("JUDGMENT & ORDER"))
    pos_list.append(text.find("J U D G M E N T"))
    pos_list.append(text.find("JUDGMENT"))
    pos_list.append(text.find("O R D E R"))
    pos_list.append(text.find("ORDER"))

    for i, p in enumerate(pos_list):

        if p != -1:
            if i == 0:
                len = 16
            elif i == 1:
                len = 15
            elif i == 2:
                len = 8
            elif i == 3:
                len = 9
            elif i == 4:
                len = 5
            pos = p + len
            break

    return pos


def get_keyword_based_preamble_end_char_offset(text):
    preamble_end_keywords = ["JUDGMENT", "ORDER", "JUDGMENT & ORDER", "COMMON ORDER",
                             "ORAL JUDGMENT"]
    preamble_end_char_offset = 0

    ### search for preamble end keywords on new lines
    for preamble_keyword in preamble_end_keywords:
        preamble_regex = re.compile(r'\n\s*[^A-Za-z0-9]{0,2}\s*' + ' {0,3}'.join([i for i in preamble_keyword]) + r'\s*[^A-Za-z0-9]{0,1}\n')
        match = re.search(preamble_regex, text)
        if match:
            preamble_end_char_offset = match.span()[1]
            break

    #### if not found then search for the keywords anywhere
    if preamble_end_char_offset == 0:
        for preamble_keyword in preamble_end_keywords:
            match = re.search(' {0,3}'.join([i for i in preamble_keyword]), text)
            if match:
                preamble_end_char_offset = match.span()[1]
                break
    return preamble_end_char_offset


def convert_upper_case_to_title(txt):
    ########### convert the uppercase words to title case for catching names in NER
    title_tokens = []
    for token in txt.split(' '):
        title_subtokens = []
        for subtoken in token.split('\n'):
            if subtoken.isupper():
                title_subtokens.append(subtoken.title())
            else:
                title_subtokens.append(subtoken)
        title_tokens.append('\n'.join(title_subtokens))
    title_txt = ' '.join(title_tokens)
    return title_txt


def guess_preamble_end(truncated_txt, nlp):
    ######### Guess the end of preamble using hueristics
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        preamble_end = 0
        max_length = 20000
        tokens = nlp.tokenizer(truncated_txt)
        if len(tokens) > max_length:
            chunks = [tokens[i:i + max_length] for i in range(0, len(tokens), max_length)]
            nlp_docs = [nlp(i.text) for i in chunks]
            truncated_doc = spacy.tokens.Doc.from_docs(nlp_docs)
        else:
            truncated_doc = nlp(truncated_txt)
    successive_preamble_pattern_breaks = 0
    preamble_patterns_breaks_theshold = 1  ####### end will be marked after these many consecutive sentences which dont match preamble pattern
    sent_list = [sent for sent in truncated_doc.sents]
    for sent_id, sent in enumerate(sent_list):
        ###### check if verb is present in the sentence
        verb_exclusions = ['reserved', 'pronounced', 'dated', 'signed']
        sent_pos_tag = [token.pos_ for token in sent if token.lower_ not in verb_exclusions]
        verb_present = 'VERB' in sent_pos_tag

        ###### check if uppercase or title case
        allowed_lowercase = ['for', 'at', 'on', 'the', 'in', 'of']
        upppercase_or_titlecase = all(
            [token.text in allowed_lowercase or token.is_upper or token.is_title or token.is_punct for token in sent if
             token.is_alpha])

        if verb_present and not upppercase_or_titlecase:
            successive_preamble_pattern_breaks += 1
            if successive_preamble_pattern_breaks > preamble_patterns_breaks_theshold:
                preamble_end = sent_list[sent_id - preamble_patterns_breaks_theshold - 1].end_char
                break
        else:
            if successive_preamble_pattern_breaks > 0 and (verb_present or not upppercase_or_titlecase):
                preamble_end = sent_list[sent_id - preamble_patterns_breaks_theshold - 1].end_char
                break
            else:
                successive_preamble_pattern_breaks = 0
    return preamble_end


def seperate_and_clean_preamble(txt, preamble_splitting_nlp):
    ########## seperate preamble from judgment text

    ######## get preamble end offset based on keywords
    keyword_preamble_end_offset = get_keyword_based_preamble_end_char_offset(txt)
    if keyword_preamble_end_offset == 0:
        preamble_end_offset = 5000  ######## if keywords not found then set arbitrarty value to reduce time for searching
    else:
        preamble_end_offset = keyword_preamble_end_offset + 200  ######## take few more characters as judge names are written after JUDGEMENT keywords
    truncated_txt = txt[:preamble_end_offset]
    guessed_preamble_end = guess_preamble_end(truncated_txt, preamble_splitting_nlp)

    if guessed_preamble_end == 0:
        preamble_end = keyword_preamble_end_offset
    else:
        preamble_end = guessed_preamble_end

    preamble_txt = txt[:preamble_end]
    title_txt = convert_upper_case_to_title(preamble_txt)
    return title_txt, preamble_end


def split_main_judgement_to_preamble_and_judgement(text, sentence_splitting_nlp, mini_batch_size=40000):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ########## Separate preamble from judgement text
        preamble_text, preamble_end = seperate_and_clean_preamble(text, sentence_splitting_nlp)

        ########## process main judgement text
        judgement_text = text[preamble_end:]
        #####  replace new lines in middle of sentence with spaces.
        judgement_text = re.sub(r'(\w[ -]*)(\n+)', r'\1 ', judgement_text)
        judgement = process_nlp_in_chunks(judgement_text, mini_batch_size, sentence_splitting_nlp)
        preamble = process_nlp_in_chunks(preamble_text, mini_batch_size, sentence_splitting_nlp)

        return preamble, judgement


def process_nlp_in_chunks(judgement_text, mini_batch_size, nlp):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        max_length = mini_batch_size
        tokens = nlp.tokenizer(judgement_text)
        if len(tokens) > max_length:
            chunks = [tokens[i:i + max_length] for i in range(0, len(tokens), max_length)]
            nlp_docs = [nlp(i.text) for i in chunks]
            judgement = spacy.tokens.Doc.from_docs(nlp_docs)
        else:
            judgement = nlp(judgement_text)
        return judgement



from hashlib import sha256
from typing import Union

import spacy
from tqdm import tqdm
from wasabi import msg

from .download import PIP_INSTALLER_URLS, install
from .sentencizer import split_main_judgement_to_preamble_and_judgement


class Data:
    def __init__(self, input_text: Union[str, list], preprocessing_nlp_model: str = 'en_core_web_trf',
                 mini_batch_size: int = 40000, use_gpu: bool = True,
                 use_cache: bool = True, verbose: bool = False, file_ids: list = []):
        """Returns object of Data class.
         It is used for common preprocessing of all the components present in this library.
        Args:
            input_text (string or list): This is where you will provide input as string or list of strings
            preprocessing_nlp_model (string): Accepts a model name of spacy as string that will be used for processing
            available models are 'en_core_web_trf', 'en_core_web_sm', 'en_core_web_md'
            mini_batch_size (int): This accepts an int as batch size for processing of a document,
             if length of document is bigger that given batch size it will be chunked and then processed.
            use_gpu (bool): Functionality to give a choice whether to use GPU for processing or not
             Setting it True doesn't ensure GPU will be utilized it need proper support libraries as mentioned in
             documentation
            use_cache (bool): Set it to true if you want to enable caching while preprocessing
            verbose (bool): Set it to if you want to see progress bar while processing happens
            file_ids (list): list of custom file ids to use with documents

            Examples::
            >>> text = 'Section 319 Cr.P.C. contemplates a situation where the evidence adduced by the prosecution for Respondent No.3-G. Sambiah on 20th June 1984'
            >>> data1 = Data(text)
            >>> data2 = Data([text,text])
            >>> data1.append(text)
            >>> data2.pop(1)
            >>> data1+text
            >>> len(data1)
            >>> text in data1
            >>> processed_data1 = [i for i in data1]
            >>> processed_data2 = data2[0]
        """
        self.__input_text__ = input_text
        self.__mini_batch_size__ = mini_batch_size
        self.__verbose__ = verbose
        self.__use_cache__ = use_cache
        self.__file_ids__ = file_ids
        if self.__use_cache__:
            self.__cache__ = {}
        if self.__file_ids__ and (len(self.__file_ids__) != len(self.__input_text__)):
            raise RuntimeError('Count of file_ids not equal to count of input text')
        if isinstance(self.__input_text__, str):
            self.__input_text__ = [self.__input_text__]
        if self.__file_ids__:
            texts = []
            ids = []
            for index, text in enumerate(self.__input_text__):
                if text.strip():
                    texts.append(text)
                    ids.append(self.__file_ids__[index])
            self.__input_text__ = texts
            self.__file_ids__ = ids
        else:
            self.__input_text__ = [text for text in self.__input_text__ if text.strip()]
        if isinstance(self.__input_text__, list) and len(self.__input_text__) >= 1 and all(
                isinstance(item, str) for item in self.__input_text__):
            pass
        else:
            raise RuntimeError('No input or wrong given, we accept input as string or list of strings')
        if preprocessing_nlp_model not in spacy.util.get_installed_models():
            msg.info(f'Installing {preprocessing_nlp_model}. This is a one time process!!')
            if PIP_INSTALLER_URLS.get(preprocessing_nlp_model) is not None and preprocessing_nlp_model in [
                'en_core_web_trf',
                'en_core_web_sm',
                'en_core_web_md']:
                install(PIP_INSTALLER_URLS[preprocessing_nlp_model])
            else:
                raise RuntimeError(
                    f'{preprocessing_nlp_model} doesn\'t exist in list of available opennyai preprocessing models')
        if use_gpu:
            try:
                if spacy.prefer_gpu():
                    msg.info(title='Pre-processing will happen on GPU!')
                else:
                    msg.info(title='Pre-processing will happen on CPU!')
                spacy.prefer_gpu()
            except:
                msg.info(title='Pre-processing will happen on CPU!')
        else:
            msg.info(title='Pre-processing will happen on CPU!')

        try:
            self.__preprocessing_nlp__ = spacy.load(preprocessing_nlp_model,
                                                    exclude=['lemmatizer', 'ner'])
        except:
            raise RuntimeError(
                f'There was an error while loading {preprocessing_nlp_model}\n To rectify try running:\n pip install -U {PIP_INSTALLER_URLS[preprocessing_nlp_model]}')

    def _clean_cache(self):
        if not self.__file_ids__:
            ids = [sha256(text.encode('utf-8')).hexdigest() for text in self.__input_text__]
        else:
            ids = self.__file_ids__
        for key in self.__cache__.keys():
            if str(key) not in ids:
                self.__cache__.pop(str(key))

    def __getitem__(self, item):
        if self.__use_cache__:
            self._clean_cache()
        to_process = self.__input_text__[item]
        if isinstance(to_process, str):
            to_process = [to_process]
        data = []
        if self.__verbose__:
            msg.info('Processing input data!!!')
        for index, text in tqdm(enumerate(to_process), disable=not self.__verbose__, total=len(to_process)):
            if self.__file_ids__:
                file_id = self.__file_ids__[item]
            else:
                file_id = sha256(text.encode('utf-8')).hexdigest()
            if self.__use_cache__ and self.__cache__.get(file_id) is not None:
                data.append(self.__cache__[file_id])
            else:
                original_text = text
                text = text.encode(encoding='ascii', errors='ignore').decode()
                preamble_doc, judgement_doc = split_main_judgement_to_preamble_and_judgement(text=text,
                                                                                             sentence_splitting_nlp=self.__preprocessing_nlp__,
                                                                                             mini_batch_size=self.__mini_batch_size__)
                preamble_doc = spacy.tokens.Doc.from_docs([i.as_doc() for i in preamble_doc.sents])
                judgement_doc = spacy.tokens.Doc.from_docs([i.as_doc() for i in judgement_doc.sents])
                if preamble_doc is None:
                    preamble_doc = self.__preprocessing_nlp__('')
                if judgement_doc is None:
                    judgement_doc = self.__preprocessing_nlp__('')
                _processed_data = {'file_id': file_id,
                                   "preamble_doc": preamble_doc,
                                   "judgement_doc": judgement_doc,
                                   "original_text": original_text}
                if self.__use_cache__:
                    self.__cache__[file_id] = _processed_data
                data.append(_processed_data)
        if len(data) == 1:
            return data[0]
        else:
            return data

    def __len__(self):
        return len(self.__input_text__)

    def __str__(self):
        if len(self.__input_text__) == 1:
            return self.__input_text__[0]
        else:
            return str(self.__input_text__)

    def __contains__(self, item):
        return True if item in self.__input_text__ else False

    def __add__(self, other):
        if isinstance(other, str):
            self.__input_text__.append(other)
        else:
            TypeError('Only str object can be added')

    def append(self, other):
        self.__add__(other)

    def pop(self, index):
        if isinstance(index, int):
            self.__input_text__.pop(index)
        else:
            TypeError("'str' object cannot be interpreted as an integer")



import numpy as np
import torch

import opennyai.rhetorical_roles.bucketing as bucketing
from opennyai.rhetorical_roles.allennlp_helper.common.util import pad_sequence_to_length


class BatchCreator:
    def __init__(self, dataset, tokenizer, labels, batch_sentence_size, max_seq_length):
        # '''dataset: Iterable over documents
        #   tokenizer: WordPiece tokenizer of BERT. If None, then it is assumed that sentences are already tokenized.
        #   labels: possible lables of the sentences
        #   max_sequence_length: Max number of tokens for each sentences. Only required if tokenizer is provided.
        # '''
        self.dataset = dataset
        self.labels = labels
        self.batch_sentence_size = batch_sentence_size
        self.max_sequence_length = max_seq_length
        self.tokenizer = tokenizer
        self.batches = None

    def get_batches(self, task_name=None):
        self.build_batches()
        result = []
        for b in self.batches:
            batch = self.batch_to_tensor(b)
            if task_name is not None:
                batch["task"] = task_name
            result.append(batch)
        return result

    def build_batches(self):
        if self.batches is None:
            def wrap_document(doc):
                return bucketing.Record(doc.get_sentence_count(), doc)

            mapped_ds = map(wrap_document, self.dataset)
            self.batches = bucketing.bucket_records(mapped_ds, self.batch_sentence_size)

        return len(self.batches)

    def get_batches_count(self):
        batches_count = self.build_batches()
        return batches_count

    def batch_to_tensor(self, b):
        # dictionary of arrays
        tensors_dict_arrays = b.to_tensor(self.document_to_sequence_example, merge_records)
        # convert to dictionary of tensors and pad the tensors
        result = {}
        for k, v in tensors_dict_arrays.items():

            if k in ["input_ids", "attention_mask"]:
                # determine the max sentence len in the batch
                max_sentence_len = -1
                for doc in v:
                    for sentence in doc:
                        max_sentence_len = max(len(sentence), max_sentence_len)
                # pad the sentences to max sentence len
                for doc in v:
                    for i, sentence in enumerate(doc):
                        doc[i] = pad_sequence_to_length(sentence, desired_length=max_sentence_len)
            if k != 'doc_name':
                result[k] = torch.tensor(v)
            else:
                result[k] = v
        return result

    def document_to_sequence_example(self, document, sentence_padding_len):

        sentences = list(document.data.sentences)
        labels = list(document.data.labels)

        # pad number of sentences
        for _ in range(len(document.data.labels), sentence_padding_len):
            sentences.append("")
            labels.append("mask")

        token_ids = []
        attention_masks = []
        label_ids = []
        for sentence, label in zip(sentences, labels):
            if self.tokenizer is None:
                # sentence already tokenized
                if isinstance(sentence, list):
                    tok_ids = sentence
                else:
                    tok_ids = [int(t) for t in sentence.split()]
            else:
                tok_ids = self.tokenizer.encode(sentence, add_special_tokens=True, max_length=self.max_sequence_length)

            attention_mask = [1] * len(tok_ids)

            # map label id
            label_id = self.labels.index(label)

            token_ids.append(tok_ids)
            attention_masks.append(attention_mask)
            label_ids.append(label_id)

        return {
            "sentence_mask": pad_sequence_to_length([1] * document.length, desired_length=sentence_padding_len),
            "input_ids": token_ids,
            "attention_mask": attention_masks,
            "label_ids": label_ids,
            "doc_name": document.data.doc_name
        }


def merge_records(merged, r):
    if merged is None:
        merged = dict()
        for k in r:
            merged[k] = []

    for k in r:
        merged[k].append(r[k])

    return merged


def one_hot(num, v):
    r = np.zeros(num, dtype=int)
    r[v] = 1
    return r



import copy
import os

import torch
import torch.nn as nn
from pytorch_transformers import BertModel, BertConfig
from torch.nn.init import xavier_uniform_

from opennyai.utils.download import CACHE_DIR
from .decoder import TransformerDecoder
from .encoder import Classifier, ExtTransformerEncoder
from .optimizers import Optimizer

EXTRACTIVE_SUMMARIZER_CACHE_PATH = os.path.join(CACHE_DIR, 'ExtractiveSummarizer'.lower())


def build_optim(args, model, checkpoint):
    """ Build optimizer """

    if checkpoint is not None:
        optim = checkpoint['optim'][0]
        saved_optimizer_state_dict = optim.optimizer.state_dict()
        optim.optimizer.load_state_dict(saved_optimizer_state_dict)
        if args.visible_gpus != '-1':
            for state in optim.optimizer.state.values():
                for k, v in state.items():
                    if torch.is_tensor(v):
                        state[k] = v.cuda()

        if (optim.method == 'adam') and (len(optim.optimizer.state) < 1):
            raise RuntimeError(
                "Error: loaded Adam optimizer from existing model" +
                " but optimizer state is empty")

    else:
        optim = Optimizer(
            args.optim, args.lr, args.max_grad_norm,
            beta1=args.beta1, beta2=args.beta2,
            decay_method='noam',
            warmup_steps=args.warmup_steps)

    optim.set_parameters(list(model.named_parameters()))

    return optim


def build_optim_bert(args, model, checkpoint):
    """ Build optimizer """

    if checkpoint is not None:
        optim = checkpoint['optims'][0]
        saved_optimizer_state_dict = optim.optimizer.state_dict()
        optim.optimizer.load_state_dict(saved_optimizer_state_dict)
        if args.visible_gpus != '-1':
            for state in optim.optimizer.state.values():
                for k, v in state.items():
                    if torch.is_tensor(v):
                        state[k] = v.cuda()

        if (optim.method == 'adam') and (len(optim.optimizer.state) < 1):
            raise RuntimeError(
                "Error: loaded Adam optimizer from existing model" +
                " but optimizer state is empty")

    else:
        optim = Optimizer(
            args.optim, args.lr_bert, args.max_grad_norm,
            beta1=args.beta1, beta2=args.beta2,
            decay_method='noam',
            warmup_steps=args.warmup_steps_bert)

    params = [(n, p) for n, p in list(model.named_parameters()) if n.startswith('bert.model')]
    optim.set_parameters(params)

    return optim


def build_optim_dec(args, model, checkpoint):
    """ Build optimizer """

    if checkpoint is not None:
        optim = checkpoint['optims'][1]
        saved_optimizer_state_dict = optim.optimizer.state_dict()
        optim.optimizer.load_state_dict(saved_optimizer_state_dict)
        if args.visible_gpus != '-1':
            for state in optim.optimizer.state.values():
                for k, v in state.items():
                    if torch.is_tensor(v):
                        state[k] = v.cuda()

        if (optim.method == 'adam') and (len(optim.optimizer.state) < 1):
            raise RuntimeError(
                "Error: loaded Adam optimizer from existing model" +
                " but optimizer state is empty")

    else:
        optim = Optimizer(
            args.optim, args.lr_dec, args.max_grad_norm,
            beta1=args.beta1, beta2=args.beta2,
            decay_method='noam',
            warmup_steps=args.warmup_steps_dec)

    params = [(n, p) for n, p in list(model.named_parameters()) if not n.startswith('bert.model')]
    optim.set_parameters(params)

    return optim


def get_generator(vocab_size, dec_hidden_size, device):
    gen_func = nn.LogSoftmax(dim=-1)
    generator = nn.Sequential(
        nn.Linear(dec_hidden_size, vocab_size),
        gen_func
    )
    generator.to(device)

    return generator


class Bert(nn.Module):
    def __init__(self, large, temp_dir=EXTRACTIVE_SUMMARIZER_CACHE_PATH, finetune=False):
        super(Bert, self).__init__()
        if (large):
            self.model = BertModel.from_pretrained('bert-large-uncased')
        else:
            self.model = BertModel.from_pretrained('bert-base-uncased')

        self.finetune = finetune

    def forward(self, x, segs, mask):
        if (self.finetune):
            top_vec, _ = self.model(x, segs, attention_mask=mask)
        else:
            self.eval()
            with torch.no_grad():
                top_vec, _ = self.model(x, segs, attention_mask=mask)
        return top_vec


class ExtSummarizer(nn.Module):
    def __init__(self, args, device, checkpoint):
        super(ExtSummarizer, self).__init__()
        self.args = args
        self.device = device
        self.bert = Bert(args.large, args.temp_dir, args.finetune_bert)

        if self.args.use_rhetorical_roles:
            self.ext_layer = ExtTransformerEncoder(self.bert.model.config.hidden_size + 16, args.ext_ff_size,
                                                   args.ext_heads,
                                                   args.ext_dropout, args.ext_layers)
        else:
            self.ext_layer = ExtTransformerEncoder(self.bert.model.config.hidden_size, args.ext_ff_size, args.ext_heads,
                                                   args.ext_dropout, args.ext_layers)
        if (args.encoder == 'baseline'):
            bert_config = BertConfig(self.bert.model.config.vocab_size, hidden_size=args.ext_hidden_size,
                                     num_hidden_layers=args.ext_layers, num_attention_heads=args.ext_heads,
                                     intermediate_size=args.ext_ff_size)
            self.bert.model = BertModel(bert_config)
            self.ext_layer = Classifier(self.bert.model.config.hidden_size)

        if (args.max_pos > 512):
            my_pos_embeddings = nn.Embedding(args.max_pos, self.bert.model.config.hidden_size)
            my_pos_embeddings.weight.data[:512] = self.bert.model.embeddings.position_embeddings.weight.data
            my_pos_embeddings.weight.data[512:] = self.bert.model.embeddings.position_embeddings.weight.data[-1][None,
                                                  :].repeat(args.max_pos - 512, 1)
            self.bert.model.embeddings.position_embeddings = my_pos_embeddings

        if checkpoint is not None:
            self.load_state_dict(checkpoint['model'], strict=True)
        else:
            if args.param_init != 0.0:
                for p in self.ext_layer.parameters():
                    p.data.uniform_(-args.param_init, args.param_init)
            if args.param_init_glorot:
                for p in self.ext_layer.parameters():
                    if p.dim() > 1:
                        xavier_uniform_(p)

        self.to(device)

    def forward(self, src, segs, clss, mask_src, mask_cls, sentence_rhetorical_roles):
        top_vec = self.bert(src, segs, mask_src)
        sents_vec = top_vec[torch.arange(top_vec.size(0)).unsqueeze(1), clss]
        sents_vec = sents_vec * mask_cls[:, :, None].float()
        if self.args.use_rhetorical_roles:
            sent_one_hot = torch.nn.functional.one_hot(sentence_rhetorical_roles, num_classes=16).float()
            sent_scores_concatenated = torch.cat((sents_vec, sent_one_hot), dim=2)
            sent_scores = self.ext_layer(sent_scores_concatenated, mask_cls).squeeze(-1)
        else:
            sent_scores = self.ext_layer(sents_vec, mask_cls).squeeze(-1)

        return sent_scores, mask_cls


class AbsSummarizer(nn.Module):
    def __init__(self, args, device, checkpoint=None, bert_from_extractive=None):
        super(AbsSummarizer, self).__init__()
        self.args = args
        self.device = device
        self.bert = Bert(args.large, args.temp_dir, args.finetune_bert)

        if bert_from_extractive is not None:
            self.bert.model.load_state_dict(
                dict([(n[11:], p) for n, p in bert_from_extractive.items() if n.startswith('bert.model')]), strict=True)

        if (args.encoder == 'baseline'):
            bert_config = BertConfig(self.bert.model.config.vocab_size, hidden_size=args.enc_hidden_size,
                                     num_hidden_layers=args.enc_layers, num_attention_heads=8,
                                     intermediate_size=args.enc_ff_size,
                                     hidden_dropout_prob=args.enc_dropout,
                                     attention_probs_dropout_prob=args.enc_dropout)
            self.bert.model = BertModel(bert_config)

        if (args.max_pos > 512):
            my_pos_embeddings = nn.Embedding(args.max_pos, self.bert.model.config.hidden_size)
            my_pos_embeddings.weight.data[:512] = self.bert.model.embeddings.position_embeddings.weight.data
            my_pos_embeddings.weight.data[512:] = self.bert.model.embeddings.position_embeddings.weight.data[-1][None,
                                                  :].repeat(args.max_pos - 512, 1)
            self.bert.model.embeddings.position_embeddings = my_pos_embeddings
        self.vocab_size = self.bert.model.config.vocab_size
        tgt_embeddings = nn.Embedding(self.vocab_size, self.bert.model.config.hidden_size, padding_idx=0)
        if (self.args.share_emb):
            tgt_embeddings.weight = copy.deepcopy(self.bert.model.embeddings.word_embeddings.weight)

        self.decoder = TransformerDecoder(
            self.args.dec_layers,
            self.args.dec_hidden_size, heads=self.args.dec_heads,
            d_ff=self.args.dec_ff_size, dropout=self.args.dec_dropout, embeddings=tgt_embeddings)

        self.generator = get_generator(self.vocab_size, self.args.dec_hidden_size, device)
        self.generator[0].weight = self.decoder.embeddings.weight

        if checkpoint is not None:
            self.load_state_dict(checkpoint['model'], strict=True)
        else:
            for module in self.decoder.modules():
                if isinstance(module, (nn.Linear, nn.Embedding)):
                    module.weight.data.normal_(mean=0.0, std=0.02)
                elif isinstance(module, nn.LayerNorm):
                    module.bias.data.zero_()
                    module.weight.data.fill_(1.0)
                if isinstance(module, nn.Linear) and module.bias is not None:
                    module.bias.data.zero_()
            for p in self.generator.parameters():
                if p.dim() > 1:
                    xavier_uniform_(p)
                else:
                    p.data.zero_()
            if (args.use_bert_emb):
                tgt_embeddings = nn.Embedding(self.vocab_size, self.bert.model.config.hidden_size, padding_idx=0)
                tgt_embeddings.weight = copy.deepcopy(self.bert.model.embeddings.word_embeddings.weight)
                self.decoder.embeddings = tgt_embeddings
                self.generator[0].weight = self.decoder.embeddings.weight

        self.to(device)

    def forward(self, src, tgt, segs, clss, mask_src, mask_tgt, mask_cls):
        top_vec = self.bert(src, segs, mask_src)
        dec_state = self.decoder.init_decoder_state(src, top_vec)
        decoder_outputs, state = self.decoder(tgt[:, :-1], top_vec, dec_state)
        return decoder_outputs, None



from setuptools import setup, find_packages

"""
Instructions for creating a release of the opennyai library.
1. Make sure your working directory is clean.
2. Make sure that you have changed the versions in "opennyai/__init__.py".
3. Create the distribution by running "python setup.py sdist" and "python setup.py bdist_wheel" in the root of the repository.
4. Check you can install the new distribution in a clean environment.
5. Upload the distribution to pypi by running
   "twine upload dist/*".
   This step will ask you for a username and password - the username & password get from [AUTHOR]
"""

setup(
    name="opennyai",
    version="0.0.14",
    url="",
    author="Aman Tiwari",
    author_email="aman.tiwari@thoughtworks.com",
    description="A SpaCy pipeline and models for NLP on indian legal text.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    keywords=[
        "law indianlegalner legalner legal ner lawtech legaltech nlp spacy SpaCy rhetorical role summarizer extractive_summarizer "],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    license="MIT",
    install_requires=[
        "torch==1.12.1",
        "typing_extensions==4.4.0",
        "torchvision<0.14.0,>=0.8.1",
        "transformers<4.16,>=4.1",
        "pytorch-transformers==1.2.0",
        "multiprocess==0.70.12.2",
        "pandas>=1.2.4,<1.3.6",
        "spacy<3.2.5,>=3.2.2",
        "spacy-transformers<1.1.6,>=1.1.4",
        "urllib3<1.26.12,>=1.26.8",
        "beautifulsoup4<4.11.0,>=4.10.0",
        "requests<2.28.2,>=2.27.1",
        "nltk<3.6.6,>=3.6",
        "tqdm>=4.63.0,<4.64.1",
        "prettytable>=3.1.1,<3.4.1",
        "scikit-learn", "pytest",
        "Levenshtein"],
    tests_require=["pytest", "pytest-cov"],
    python_requires=">=3.7.0",
)



import copy
import math

import torch
from transformers import BertModel

from opennyai.rhetorical_roles.allennlp_helper.common.util import pad_sequence_to_length
from opennyai.rhetorical_roles.allennlp_helper.modules.conditional_random_field.conditional_random_field import \
    ConditionalRandomField
from opennyai.rhetorical_roles.allennlp_helper.modules.seq2seq_encoders.pytorch_seq2seq_wrapper import \
    PytorchSeq2SeqWrapper
from opennyai.rhetorical_roles.allennlp_helper.nn.util import masked_softmax


class CRFOutputLayer(torch.nn.Module):
    ''' CRF output layer consisting of a linear layer and a CRF. '''

    def __init__(self, in_dim, num_labels):
        super(CRFOutputLayer, self).__init__()
        self.num_labels = num_labels
        self.classifier = torch.nn.Linear(in_dim, self.num_labels)
        self.crf = ConditionalRandomField(self.num_labels)

    def forward(self, x, mask, labels=None):
        ''' x: shape: batch, max_sequence, in_dim
            mask: shape: batch, max_sequence
            labels: shape: batch, max_sequence
        '''

        batch_size, max_sequence, in_dim = x.shape

        logits = self.classifier(x)
        outputs = {}
        if labels is not None:
            log_likelihood = self.crf(logits, labels, mask)
            loss = -log_likelihood
            outputs["loss"] = loss
        else:
            best_paths = self.crf.viterbi_tags(logits, mask)
            predicted_label = [x for x, y in best_paths]
            predicted_label = [pad_sequence_to_length(x, desired_length=max_sequence) for x in predicted_label]
            predicted_label = torch.tensor(predicted_label)
            outputs["predicted_label"] = predicted_label

            # log_denominator = self.crf._input_likelihood(logits, mask)
            # log_numerator = self.crf._joint_likelihood(logits, predicted_label, mask)
            # log_likelihood = log_numerator - log_denominator
            # outputs["log_likelihood"] = log_likelihood

        return outputs


class CRFPerTaskOutputLayer(torch.nn.Module):
    ''' CRF output layer consisting of a linear layer and a CRF. '''

    def __init__(self, in_dim, tasks):
        super(CRFPerTaskOutputLayer, self).__init__()

        self.per_task_output = torch.nn.ModuleDict()
        for task in tasks:
            self.per_task_output[task.task_name] = CRFOutputLayer(in_dim=in_dim, num_labels=len(task.labels))

    def forward(self, task, x, mask, labels=None, output_all_tasks=False):
        ''' x: shape: batch, max_sequence, in_dim
            mask: shape: batch, max_sequence
            labels: shape: batch, max_sequence
        '''
        output = self.per_task_output[task](x, mask, labels)
        if output_all_tasks:
            output["task_outputs"] = []
            assert labels is None
            for t, task_output in self.per_task_output.items():
                task_result = task_output(x, mask)
                task_result["task"] = t
                output["task_outputs"].append(task_result)
        return output

    def to_device(self, device1, device2):
        self.task_to_device = dict()
        for index, task in enumerate(self.per_task_output.keys()):
            if index % 2 == 0:
                self.task_to_device[task] = device1
                self.per_task_output[task].to(device1)
            else:
                self.task_to_device[task] = device2
                self.per_task_output[task].to(device2)

    def get_device(self, task):
        return self.task_to_device[task]


class AttentionPooling(torch.nn.Module):
    def __init__(self, in_features, dimension_context_vector_u=200, number_context_vectors=5):
        super(AttentionPooling, self).__init__()
        self.dimension_context_vector_u = dimension_context_vector_u
        self.number_context_vectors = number_context_vectors
        self.linear1 = torch.nn.Linear(in_features=in_features, out_features=self.dimension_context_vector_u, bias=True)
        self.linear2 = torch.nn.Linear(in_features=self.dimension_context_vector_u,
                                       out_features=self.number_context_vectors, bias=False)

        self.output_dim = self.number_context_vectors * in_features

    def forward(self, tokens, mask):
        # shape tokens: (batch_size, tokens, in_features)

        # compute the weights
        # shape tokens: (batch_size, tokens, dimension_context_vector_u)
        a = self.linear1(tokens)
        a = torch.tanh(a)
        # shape (batch_size, tokens, number_context_vectors)
        a = self.linear2(a)
        # shape (batch_size, number_context_vectors, tokens)
        a = a.transpose(1, 2)
        a = masked_softmax(a, mask)

        # calculate weighted sum
        s = torch.bmm(a, tokens)
        s = s.view(tokens.shape[0], -1)
        return s


class BertTokenEmbedder(torch.nn.Module):
    def __init__(self, config):
        super(BertTokenEmbedder, self).__init__()
        self.bert = BertModel.from_pretrained(config["bert_model"])
        # state_dict_1 = self.bert.state_dict()
        # state_dict_2 = torch.load('/home/astha_agarwal/model/pytorch_model.bin')
        # for name2 in state_dict_2.keys():
        #    for name1 in state_dict_1.keys():
        #        temp_name = copy.deepcopy(name2)
        #       if temp_name.replace("bert.", '') == name1:
        #            state_dict_1[name1] = state_dict_2[name2]

        # self.bert.load_state_dict(state_dict_1,strict=False)

        self.bert_trainable = config["bert_trainable"]
        self.bert_hidden_size = self.bert.config.hidden_size
        self.cacheable_tasks = config["cacheable_tasks"]
        for param in self.bert.parameters():
            param.requires_grad = self.bert_trainable

    def forward(self, batch):
        documents, sentences, tokens = batch["input_ids"].shape

        if "bert_embeddings" in batch:
            return batch["bert_embeddings"]

        attention_mask = batch["attention_mask"].view(-1, tokens)

        # input_ids = batch["input_ids"].view(-1, tokens)
        # outputs = self.bert(input_ids=input_ids, attention_mask=attention_mask)
        # # shape (documents*sentences, tokens, 768)
        # bert_embeddings = outputs[0]

        #### break the large judgements into sentences chunk of given size. Do this while inference
        chunk_size = 1024
        input_ids = batch["input_ids"].view(-1, tokens)
        chunk_cnt = int(math.ceil(input_ids.shape[0] / chunk_size))
        input_ids_chunk_list = torch.chunk(input_ids, chunk_cnt)

        attention_mask_chunk_list = torch.chunk(attention_mask, chunk_cnt)
        outputs = []
        for input_ids, attention_mask in zip(input_ids_chunk_list, attention_mask_chunk_list):
            with torch.no_grad():
                output = self.bert(input_ids=input_ids, attention_mask=attention_mask)
                output = output[0]
                # output = output[0].to('cpu')
            outputs.append(copy.deepcopy(output))
            torch.cuda.empty_cache()

        bert_embeddings = torch.cat(tuple(outputs))  # .to('cuda')

        if not self.bert_trainable and batch["task"] in self.cacheable_tasks:
            # cache the embeddings of BERT if it is not fine-tuned
            # to save GPU memory put the values on CPU
            batch["bert_embeddings"] = bert_embeddings.to("cpu")

        return bert_embeddings


class BertHSLN(torch.nn.Module):
    '''
    Model for Baseline, Sequential Transfer Learning and Multitask-Learning with all layers shared (except output layer).
    '''

    def __init__(self, config, tasks):
        super(BertHSLN, self).__init__()

        self.bert = BertTokenEmbedder(config)

        # Jin et al. uses DROPOUT WITH EXPECTATION-LINEAR REGULARIZATION (see Ma et al. 2016),
        # we use instead default dropout
        self.dropout = torch.nn.Dropout(config["dropout"])

        self.generic_output_layer = config.get("generic_output_layer")

        self.lstm_hidden_size = config["word_lstm_hs"]

        self.word_lstm = PytorchSeq2SeqWrapper(torch.nn.LSTM(input_size=self.bert.bert_hidden_size,
                                                             hidden_size=self.lstm_hidden_size,
                                                             num_layers=1, batch_first=True, bidirectional=True))

        self.attention_pooling = AttentionPooling(2 * self.lstm_hidden_size,
                                                  dimension_context_vector_u=config["att_pooling_dim_ctx"],
                                                  number_context_vectors=config["att_pooling_num_ctx"])

        self.init_sentence_enriching(config, tasks)

        self.reinit_output_layer(tasks, config)

    def init_sentence_enriching(self, config, tasks):
        input_dim = self.attention_pooling.output_dim
        # print(f"Attention pooling dim: {input_dim}")
        self.sentence_lstm = PytorchSeq2SeqWrapper(torch.nn.LSTM(input_size=input_dim,
                                                                 hidden_size=self.lstm_hidden_size,
                                                                 num_layers=1, batch_first=True, bidirectional=True))

    def reinit_output_layer(self, tasks, config):
        if config.get("without_context_enriching_transfer"):
            self.init_sentence_enriching(config, tasks)
        input_dim = self.lstm_hidden_size * 2

        if self.generic_output_layer:
            self.crf = CRFOutputLayer(in_dim=input_dim, num_labels=len(tasks[0].labels))
        else:
            self.crf = CRFPerTaskOutputLayer(input_dim, tasks)

    def forward(self, batch, labels=None, output_all_tasks=False):

        documents, sentences, tokens = batch["input_ids"].shape

        # shape (documents*sentences, tokens, 768)
        bert_embeddings = self.bert(batch)

        # in Jin et al. only here dropout
        bert_embeddings = self.dropout(bert_embeddings)

        tokens_mask = batch["attention_mask"].view(-1, tokens)
        # shape (documents*sentences, tokens, 2*lstm_hidden_size)
        bert_embeddings_encoded = self.word_lstm(bert_embeddings, tokens_mask)

        # shape (documents*sentences, pooling_out)
        # sentence_embeddings = torch.mean(bert_embeddings_encoded, dim=1)
        sentence_embeddings = self.attention_pooling(bert_embeddings_encoded, tokens_mask)
        # shape: (documents, sentences, pooling_out)
        sentence_embeddings = sentence_embeddings.view(documents, sentences, -1)
        # in Jin et al. only here dropout
        sentence_embeddings = self.dropout(sentence_embeddings)

        sentence_mask = batch["sentence_mask"]

        # shape: (documents, sentence, 2*lstm_hidden_size)
        sentence_embeddings_encoded = self.sentence_lstm(sentence_embeddings, sentence_mask)
        # in Jin et al. only here dropout
        sentence_embeddings_encoded = self.dropout(sentence_embeddings_encoded)

        if self.generic_output_layer:
            output = self.crf(sentence_embeddings_encoded, sentence_mask, labels)
        else:
            output = self.crf(batch["task"], sentence_embeddings_encoded, sentence_mask, labels, output_all_tasks)

        return output


class BertHSLNMultiSeparateLayers(torch.nn.Module):
    '''
    Model Multi-Task Learning, where only certail layers are shared.
    This class is necessary to separate the model on two GPUs.
    '''

    def __init__(self, config, tasks):
        super(BertHSLNMultiSeparateLayers, self).__init__()

        self.bert = BertTokenEmbedder(config)

        # Jin et al. uses DROPOUT WITH EXPECTATION-LINEAR REGULARIZATION (see Ma et al. 2016),
        # we use instead default dropout
        self.dropout = torch.nn.Dropout(config["dropout"])

        self.lstm_hidden_size = config["word_lstm_hs"]

        self.word_lstm = PytorchSeq2SeqWrapper(torch.nn.LSTM(input_size=self.bert.bert_hidden_size,
                                                             hidden_size=self.lstm_hidden_size,
                                                             num_layers=1, batch_first=True, bidirectional=True))

        self.attention_pooling = PerTaskGroupWrapper(
            task_groups=config["attention_groups"],
            create_module_func=lambda g:
            AttentionPooling(2 * self.lstm_hidden_size,
                             dimension_context_vector_u=config["att_pooling_dim_ctx"],
                             number_context_vectors=config["att_pooling_num_ctx"])
        )

        attention_pooling_output_dim = next(iter(self.attention_pooling.per_task_mod.values())).output_dim
        self.sentence_lstm = PerTaskGroupWrapper(
            task_groups=config["context_enriching_groups"],
            create_module_func=lambda g:
            PytorchSeq2SeqWrapper(torch.nn.LSTM(input_size=attention_pooling_output_dim,
                                                hidden_size=self.lstm_hidden_size,
                                                num_layers=1, batch_first=True, bidirectional=True))
        )

        self.crf = CRFPerTaskGroupOutputLayer(self.lstm_hidden_size * 2, tasks, config["output_groups"])

    def to_device(self, device1, device2):
        self.bert.to(device1)
        self.word_lstm.to(device1)
        self.attention_pooling.to_device(device1, device2)
        self.sentence_lstm.to_device(device1, device2)
        self.crf.to_device(device1, device2)
        self.device1 = device1
        self.device2 = device2

    def forward(self, batch, labels=None, output_all_tasks=False):
        task_name = batch["task"]
        documents, sentences, tokens = batch["input_ids"].shape

        # shape (documents*sentences, tokens, 768)
        bert_embeddings = self.bert(batch)

        # in Jin et al. only here dropout
        bert_embeddings = self.dropout(bert_embeddings)

        tokens_mask = batch["attention_mask"].view(-1, tokens)
        # shape (documents*sentences, tokens, 2*lstm_hidden_size)
        bert_embeddings_encoded = self.word_lstm(bert_embeddings, tokens_mask)

        # shape (documents*sentences, pooling_out)
        # sentence_embeddings = torch.mean(bert_embeddings_encoded, dim=1)
        device = self.attention_pooling.get_device(task_name)
        sentence_embeddings = self.attention_pooling(task_name, bert_embeddings_encoded.to(device),
                                                     tokens_mask.to(device))
        # shape: (documents, sentences, pooling_out)
        sentence_embeddings = sentence_embeddings.view(documents, sentences, -1)
        # in Jin et al. only here dropout
        sentence_embeddings = self.dropout(sentence_embeddings)

        sentence_mask = batch["sentence_mask"]
        # shape: (documents, sentence, 2*lstm_hidden_size)
        device = self.sentence_lstm.get_device(task_name)
        sentence_embeddings_encoded = self.sentence_lstm(task_name, sentence_embeddings.to(device),
                                                         sentence_mask.to(device))
        # in Jin et al. only here dropout
        sentence_embeddings_encoded = self.dropout(sentence_embeddings_encoded)

        device = self.crf.get_device(task_name)
        if labels is not None:
            labels = labels.to(device)

        output = self.crf(task_name, sentence_embeddings_encoded.to(device), sentence_mask.to(device), labels,
                          output_all_tasks)

        return output


class CRFPerTaskGroupOutputLayer(torch.nn.Module):
    ''' CRF output layer consisting of a linear layer and a CRF. '''

    def __init__(self, in_dim, tasks, task_groups):
        super(CRFPerTaskGroupOutputLayer, self).__init__()

        def get_task(name):
            for t in tasks:
                if t.task_name == name:
                    return t

        self.crf = PerTaskGroupWrapper(
            task_groups=task_groups,
            create_module_func=lambda g:
            # we assume same labels per group
            CRFOutputLayer(in_dim=in_dim, num_labels=len(get_task(g[0]).labels))
        )
        self.all_tasks = [t for t in [g for g in task_groups]]

    def forward(self, task, x, mask, labels=None, output_all_tasks=False):
        ''' x: shape: batch, max_sequence, in_dim
            mask: shape: batch, max_sequence
            labels: shape: batch, max_sequence
        '''
        output = self.crf(task, x, mask, labels)
        if output_all_tasks:
            output["task_outputs"] = []
            assert labels is None
            for task in self.self.all_tasks:
                task_result = self.crf(task, x, mask, labels)
                task_result["task"] = task
                output["task_outputs"].append(task_result)
        return output

    def to_device(self, device1, device2):
        self.crf.to_device(device1, device2)

    def get_device(self, task):
        return self.crf.get_device(task)


class PerTaskGroupWrapper(torch.nn.Module):
    def __init__(self, task_groups, create_module_func):
        super(PerTaskGroupWrapper, self).__init__()

        self.per_task_mod = torch.nn.ModuleDict()
        for g in task_groups:
            mod = create_module_func(g)
            for t in g:
                self.per_task_mod[t] = mod

        self.task_groups = task_groups

    def forward(self, task_name, *args):
        mod = self.per_task_mod[task_name]
        return mod(*args)

    def to_device(self, device1, device2):
        self.task_to_device = dict()
        for index, tasks in enumerate(self.task_groups):
            for task in tasks:
                if index % 2 == 0:
                    self.task_to_device[task] = device1
                    self.per_task_mod[task].to(device1)
                else:
                    self.task_to_device[task] = device2
                    self.per_task_mod[task].to(device2)

    def get_device(self, task):
        return self.task_to_device[task]



import copy

from ..prepro.data_builder import BertData, greedy_selection


def preprocess_for_summarization(input_data: dict, bert_tokenizer, tokenizer):
    """
    input_data={"id":file_name,data{"text"},"annotations":[....]}
    :param bert_tokenizer:
    :param tokenizer:
    :param input_data:
    :return bert_preprocessed_data:
    """
    max_bert_tokens_per_chunk = 510

    source_chunk_id = 0
    doc_data = {"src": [], 'tgt': [], "src_rhetorical_roles": [], 'source_filename': input_data['id'],
                'sentence_id': [], 'src_chunk_id': source_chunk_id}
    doc_data_list = []

    for index, value in enumerate(input_data['annotations']):
        value = value
        if not value['labels'][0] in ['NONE', 'PREAMBLE']:
            tokenized = bert_tokenizer.tokenize(value['text'])
            sent_tokens = [token.text for token in tokenizer(value['text'])]
            if (sum([len(bert_tokenizer.tokenize(' '.join(i))) + 2 for i in doc_data['src']]) + len(
                    tokenized)) <= max_bert_tokens_per_chunk:
                doc_data['src'].append(sent_tokens)
                doc_data['src_rhetorical_roles'].append(value['labels'][0])
                doc_data['sentence_id'].append(value['id'])
            elif len(
                    tokenized) > max_bert_tokens_per_chunk:
                if doc_data['src']:
                    doc_data_list.append(copy.deepcopy(doc_data))
                    source_chunk_id += 1
                    doc_data = {"src": [], 'tgt': [], "src_rhetorical_roles": [],
                                'source_filename': input_data['id'],
                                'sentence_id': [],
                                'src_chunk_id': source_chunk_id}
                tokens_list = [tokenized[i:i + max_bert_tokens_per_chunk] for i in
                               range(0, len(tokenized), max_bert_tokens_per_chunk - 0)]
                if len(tokens_list[-1]) < 100:
                    tokens_list = tokens_list[:-1]
                misc_sentence_id = float(value['id'].split('_')[-1])
                for _ in tokens_list:
                    misc_sentence_id += 0.01
                    sent_tokens = bert_tokenizer.convert_tokens_to_string(_).split(
                        ' ')  # [token.text for token in self.tokenizer(value['text'])]
                    doc_data['src'].append(sent_tokens)
                    doc_data['src_rhetorical_roles'].append(value['labels'][0])
                    doc_data['sentence_id'].append(misc_sentence_id)
                    doc_data_list.append(copy.deepcopy(doc_data))
                    source_chunk_id += 1
                    doc_data = {"src": [], 'tgt': [], "src_rhetorical_roles": [],
                                'source_filename': input_data['id'],
                                'sentence_id': [],
                                'src_chunk_id': source_chunk_id}
            else:
                doc_data_list.append(copy.deepcopy(doc_data))
                source_chunk_id += 1
                doc_data = {"src": [], 'tgt': [], "src_rhetorical_roles": [],
                            'source_filename': input_data['id'],
                            'sentence_id': [],
                            'src_chunk_id': source_chunk_id}
                doc_data['src'].append(sent_tokens)
                doc_data['src_rhetorical_roles'].append(value['labels'][0])
                doc_data['sentence_id'].append(value['id'])

    if doc_data['src']:
        doc_data_list.append(copy.deepcopy(doc_data))

    return doc_data_list


def format_to_bert(bert_preprocessed_data, preprocessing_args):
    bert = BertData(preprocessing_args)
    datasets = []
    for d in bert_preprocessed_data:
        source, tgt = d['src'], d['tgt']
        f_name_chunk_id = '___'.join([d['source_filename'], str(d['src_chunk_id']), str(d['sentence_id'])])
        sent_labels = greedy_selection(source[:preprocessing_args.max_src_nsents], tgt, len(tgt))
        sent_rhetorical_roles = d['src_rhetorical_roles']
        if (preprocessing_args.lower):
            source = [' '.join(s).lower().split() for s in source]
            tgt = [' '.join(s).lower().split() for s in tgt]
        b_data = bert.preprocess(source, tgt, sent_labels, sent_rhetorical_roles,
                                 use_bert_basic_tokenizer=preprocessing_args.use_bert_basic_tokenizer,
                                 is_test=True)
        # b_data = bert.preprocess(source, tgt, sent_labels, use_bert_basic_tokenizer=args.use_bert_basic_tokenizer)

        if (b_data is None):
            continue
        src_subtoken_idxs, sent_labels, sent_rr, tgt_subtoken_idxs, segments_ids, cls_ids, src_txt, tgt_txt = b_data
        b_data_dict = {"src": src_subtoken_idxs, "tgt": tgt_subtoken_idxs,
                       "src_sent_labels": sent_labels, "sentence_rhetorical_roles": sent_rr, "segs": segments_ids,
                       'clss': cls_ids,
                       'src_txt': src_txt, "tgt_txt": tgt_txt, 'unique_id': f_name_chunk_id}
        datasets.append(b_data_dict)
    return datasets



from opennyai import ner, utils
from opennyai.pipeline import Pipeline
from opennyai.rhetorical_roles.rhetorical_roles import RhetoricalRolePredictor
from opennyai.summarizer.ExtractiveSummarizer import ExtractiveSummarizer

__version__ = "{0}.{1}.{2}".format("0", "0", "14")
__author__ = 'OpenNyAI Team'
__credits__ = 'Opennyai'



import copy

import spacy
from tqdm import tqdm
from wasabi import msg

from opennyai.utils.download import install, PIP_INSTALLER_URLS
from .entity_recognizer_utils import extract_entities_from_judgment_text
from .postprocessing_utils import precedent_coref_resol, other_person_coref_res, pro_statute_coref_resol, \
    remove_overlapping_entities


class InLegalNER:
    def __init__(self, model_name='en_legal_ner_trf', use_gpu=True):
        """Returns object of InLegalNER class.
         It is used for loading InLegalNER model in memory.
        Args:
            model_name (string): Accepts a model name of spacy as InLegalNER that will be used for NER inference
            available models are 'en_legal_ner_trf', 'en_legal_ner_sm'
            use_gpu (bool): Functionality to give a choice whether to use GPU for inference or not
             Setting it True doesn't ensure GPU will be utilized it need proper support libraries as mentioned in
             documentation
        """
        if model_name not in spacy.util.get_installed_models():
            msg.info(f'Installing {model_name}. This is a one time process!!')
            if PIP_INSTALLER_URLS.get(model_name) is not None:
                install(PIP_INSTALLER_URLS[model_name])
            else:
                raise RuntimeError(f'{model_name} doesn\'t exist in list of available opennyai ner models')
        if use_gpu:
            try:
                if spacy.prefer_gpu():
                    msg.info(title='NER will run on GPU!')
                else:
                    msg.info(title='NER will run on CPU!')
                spacy.prefer_gpu()
            except:
                msg.info(title='NER will run on CPU!')
        else:
            msg.info(title='NER will run on CPU!')
        self.model_name = model_name
        self.nlp = spacy.load(self.model_name)

    def __call__(self, data, do_sentence_level=True, do_postprocess=True, mini_batch_size=40000, verbose=False,
                 statute_shortforms_path=''):
        """Returns doc of InLegalNER nlp.
         It is used for doing inference on input doc using InLegalNER model in memory.
        Args:
            data (object of Data class): Data class object containing input docs
            do_sentence_level (bool): To perform inference at sentence level or not, at sentence level it better accuracy
            do_postprocess (bool): To perform post-processing over processed doc
            mini_batch_size (int): This accepts an int as batch size for processing of a document,
            if length of document is bigger that given batch size it will be chunked and then processed.
            verbose (bool): Set it to if you want to see progress bar while processing happens
        Note:
            do_postprocess is depends on do_sentence_level so for doing postprocessing doing sentence level inference
            is mandatory.
        """
        processed_data = []
        if verbose:
            msg.info('Processing documents with Legal NER!!!')
        for to_process in tqdm(data, disable=not verbose):
            nlp_doc = extract_entities_from_judgment_text(to_process=to_process, legal_nlp=self.nlp,
                                                          do_sentence_level=do_sentence_level,
                                                          mini_batch_size=mini_batch_size)
            nlp_doc.user_data['doc_id'] = to_process['file_id']
            nlp_doc.user_data['original_text'] = to_process['original_text']
            nlp_doc.user_data['preamble_end_char_offset'] = len(to_process['preamble_doc'].text)
            if do_sentence_level and do_postprocess:
                all_entities = None
                postprocessing_success = True
                precedent_clusters = precedent_coref_resol(nlp_doc)

                other_person_entites = other_person_coref_res(nlp_doc)

                pro_sta_clusters, stat_clusters = pro_statute_coref_resol(nlp_doc, statute_shortforms_path)
                if pro_sta_clusters:
                    all_entities = remove_overlapping_entities(nlp_doc.ents, pro_sta_clusters)

                    all_entities.extend(other_person_entites)
                elif pro_sta_clusters is None and stat_clusters is None:
                    postprocessing_success = False

                if all_entities:
                    nlp_doc.ents = all_entities
                else:
                    ents = list(nlp_doc.ents)
                    ents.extend(other_person_entites)
                    ents = spacy.util.filter_spans(tuple(ents))
                    nlp_doc.ents = ents

                if precedent_clusters:
                    nlp_doc.user_data['precedent_clusters'] = precedent_clusters
                elif precedent_clusters is None:
                    postprocessing_success = False

                if pro_sta_clusters:
                    nlp_doc.user_data['provision_statute_pairs'] = pro_sta_clusters
                elif pro_sta_clusters is None:
                    postprocessing_success = False

                if stat_clusters:
                    nlp_doc.user_data['statute_clusters'] = stat_clusters
                elif stat_clusters is None:
                    postprocessing_success = False

                if not postprocessing_success:
                    msg.warn(
                        f'''There was some issue while performing postprocessing for doc id {to_process['file_id']}. 
                        Some of postprocessing info may be absent because of this in doc.''')
            processed_data.append(nlp_doc)
        if len(processed_data) == 1:
            return processed_data[0]
        else:
            return processed_data



import json
import os

from opennyai import Pipeline
from opennyai.ner import get_json_from_spacy_doc
from opennyai.utils import Data
from ..utils import reset_ids, reset_sent_scores

# Execute pipeline on the text
saved_data_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
text1 = open(os.path.join(saved_data_path, 'examples/judgement_texts/72703592.txt')).read()
text2 = open(os.path.join(saved_data_path, 'examples/judgement_texts/811682.txt')).read()
texts_to_process = [text1, text2]
data = Data(texts_to_process)
pipeline = Pipeline(components=['NER', 'Rhetorical_Role', 'Summarizer'], use_gpu=True)
results = pipeline(data)

# load saved results
ner_model_output_text1 = reset_ids(
    json.load(open(os.path.join(saved_data_path, 'examples/ner_output_files/72703592.json'))))
rr_model_output_text1 = reset_sent_scores(reset_ids(
    json.load(open(os.path.join(saved_data_path, 'examples/rr_output_files/72703592.json')))))
summarizer_model_output_text1 = json.load(
    open(os.path.join(saved_data_path, 'examples/summarizer_output_files/72703592.json')))

ner_model_output_text2 = reset_ids(
    json.load(open(os.path.join(saved_data_path, 'examples/ner_output_files/811682.json'))))
rr_model_output_text2 = reset_sent_scores(reset_ids(
    json.load(open(os.path.join(saved_data_path, 'examples/rr_output_files/811682.json')))))
summarizer_model_output_text2 = json.load(
    open(os.path.join(saved_data_path, 'examples/summarizer_output_files/811682.json')))

pipeline_output_text1 = reset_sent_scores(reset_ids(
    json.load(open(os.path.join(saved_data_path, 'examples/pipeline_output_files/72703592.json')))))
pipeline_output_text2 = reset_sent_scores(reset_ids(
    json.load(open(os.path.join(saved_data_path, 'examples/pipeline_output_files/811682.json')))))


def test_ner_output():
    ner_output_valid = True
    output1 = reset_ids(get_json_from_spacy_doc(pipeline._ner_model_output[0]))
    output2 = reset_ids(get_json_from_spacy_doc(pipeline._ner_model_output[1]))
    if output1 != ner_model_output_text1 or output2 != ner_model_output_text2:
        ner_output_valid = False
    assert ner_output_valid


def test_rr_output():
    rr_output_valid = True
    output1 = reset_sent_scores(reset_ids(pipeline._rr_model_output[0]))
    output2 = reset_sent_scores(reset_ids(pipeline._rr_model_output[1]))
    if output1 != rr_model_output_text1 or output2 != rr_model_output_text2:
        rr_output_valid = False
    assert rr_output_valid


def test_summarizer_output():
    summarizer_output_valid = True
    output1 = pipeline._summarizer_model_output[0]
    output2 = pipeline._summarizer_model_output[1]
    if output1 != summarizer_model_output_text1 or output2 != summarizer_model_output_text2:
        summarizer_output_valid = False
    assert summarizer_output_valid


def test_pipeline_output():
    pipeline_output_valid = True
    output1 = reset_sent_scores(reset_ids(results[0]))
    output2 = reset_sent_scores(reset_ids(results[1]))
    if output1 != pipeline_output_text1 or output2 != pipeline_output_text2:
        pipeline_output_valid = False
    assert pipeline_output_valid
